# Licensed under a 3-clause BSD style license - see LICENSE.rst
from pathlib import Path

import numpy as np
import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy import wcs as fitswcs
from astropy.io import fits
from astropy.modeling import models
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose

from gwcs import utils as gwutils
from gwcs.utils import UnsupportedProjectionError

from . import data

data_path = Path(data.__file__).parent.absolute()


def test_wrong_projcode():
    ctype = {"CTYPE": ["RA---TAM", "DEC--TAM"]}
    with pytest.raises(UnsupportedProjectionError):
        gwutils.get_projcode(ctype)


def test_wrong_projcode2():
    with pytest.raises(UnsupportedProjectionError):
        gwutils.create_projection_transform("TAM")


def test_fits_transform():
    hdr = fits.Header.fromfile(data_path / "simple_wcs2.hdr")
    gw1 = gwutils.make_fitswcs_transform(hdr)
    w1 = fitswcs.WCS(hdr)
    assert_allclose(gw1(1, 2), w1.wcs_pix2world(1, 2, 0), atol=10**-8)


def test_lon_pole():
    tan = models.Pix2Sky_TAN()
    car = models.Pix2Sky_CAR()
    azp = models.Pix2Sky_AZP(mu=-1.35, gamma=25.8458)
    sky_positive_lat = coord.SkyCoord(3 * u.deg, 1 * u.deg)
    sky_negative_lat = coord.SkyCoord(3 * u.deg, -1 * u.deg)
    assert_quantity_allclose(
        gwutils._compute_lon_pole(sky_positive_lat, tan), 180 * u.deg
    )
    assert_quantity_allclose(
        gwutils._compute_lon_pole(sky_negative_lat, tan), 180 * u.deg
    )
    assert_quantity_allclose(
        gwutils._compute_lon_pole(sky_positive_lat, car), 0 * u.deg
    )
    assert_quantity_allclose(
        gwutils._compute_lon_pole(sky_negative_lat, car), 180 * u.deg
    )
    assert_quantity_allclose(
        gwutils._compute_lon_pole((0, 0.34 * u.rad), tan), 180 * u.deg
    )
    assert_quantity_allclose(
        gwutils._compute_lon_pole((1 * u.rad, 0.34 * u.rad), azp), 180 * u.deg
    )
    assert_allclose(gwutils._compute_lon_pole((1, -34), tan), 180)
    assert_allclose(gwutils._compute_lon_pole((1, -90), tan), 180)
    assert_allclose(gwutils._compute_lon_pole((1, 90), tan), 180)


def test_unknown_ctype():
    wcsinfo = {
        "CDELT": np.array([3.61111098e-05, 3.61111098e-05, 2.49999994e-03]),
        "CRPIX": np.array([17.0, 16.0, 1.0]),
        "CRVAL": np.array([4.49999564e01, 1.72786731e-04, 4.84631542e00]),
        "CTYPE": np.array(["MRSAL1A", "MRSBE1A", "WAVE"]),
        "CUNIT": np.array([u.Unit("deg"), u.Unit("deg"), u.Unit("um")], dtype=object),
        "PC": np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
        "WCSAXES": 3,
        "has_cd": False,
    }
    transform = gwutils.make_fitswcs_transform(wcsinfo)
    x = np.linspace(-5, 7, 10)
    y = np.linspace(-5, 7, 10)
    expected = (
        np.array(
            [
                -0.00075833,
                -0.00071019,
                -0.00066204,
                -0.00061389,
                -0.00056574,
                -0.00051759,
                -0.00046944,
                -0.0004213,
                -0.00037315,
                -0.000325,
            ]
        ),
        np.array(
            [
                -0.00072222,
                -0.00067407,
                -0.00062593,
                -0.00057778,
                -0.00052963,
                -0.00048148,
                -0.00043333,
                -0.00038519,
                -0.00033704,
                -0.00028889,
            ]
        ),
    )
    a, b = transform(x, y)
    assert_allclose(a, expected[0], atol=10**-8)
    assert_allclose(b, expected[1], atol=10**-8)


def test_get_axes():
    wcsinfo = {"CTYPE": np.array(["MRSAL1A", "MRSBE1A", "WAVE"])}
    cel, spec, other = gwutils.get_axes(wcsinfo)
    assert not cel
    assert spec == [2]
    assert other == [0, 1]
    wcsinfo = {"CTYPE": np.array(["RA---TAN", "WAVE", "DEC--TAN"])}
    cel, spec, other = gwutils.get_axes(wcsinfo)
    assert cel == [0, 2]
    assert spec == [1]
    assert not other


def test_get_values():
    args = 2 * u.cm
    units = (u.m,)
    res = gwutils.get_values(units, args)
    assert res == [0.02]

    res = gwutils.get_values(None, args)
    assert res == [2]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Models for generating FITS WCS standard transforms.
"""

from astropy.modeling.core import Model
from astropy.modeling.models import (
    AffineTransformation2D,
    RotateNative2Celestial,
    Scale,
    Shift,
)
from astropy.modeling.parameters import Parameter

from .utils import _compute_lon_pole

__all__ = [
    "FITSImagingWCSTransform",
]


class FITSImagingWCSTransform(Model):
    """
    FITS WCS standard imaging transform.

    Parameters
    ----------
    projection : `~astropy.modeling.projections.Projection`
      Celestial projection.
    crpix : tuple or iterable of size 2
      Pixel coordinate of the reference point. Default is [0, 0]
    crval : tuple or iterable of size 2
      Celestial longitude and latitude of the fiducial point in deg. Default is [0, 0]
    cdelt : tuple or iterable of size 2
      Coordinate scale factors. Default is [1, 1]
    pc : ndarray of shape (2,2)
      Linear transformation matrix. Default is [[1,0][0, 1]]


    Returns
    -------
    transform : `~astropy.modeling.core.CompoundModel`
      The transform between a detector and a celestial coordinate frames.

    Examples
    --------
    >>> from gwcs import FITSImagingWCSTransform
    >>> from astropy.modeling.models import Pix2Sky_Gnomonic
    >>> tan = Pix2Sky_Gnomonic()
    >>> fwcs = FITSImagingWCSTransform(tan, crpix=[10,10], crval=[5.36, -72.5])
    >>> print(fwcs)
    Model: FITSImagingWCSTransform
    Inputs: ('x', 'y')
    Outputs: ('lon', 'lat')
    Model set size: 1
    Parameters:
       crpix         crval       cdelt        pc
    ------------ ------------- ---------- ----------
    10.0 .. 10.0 5.36 .. -72.5 1.0 .. 1.0 1.0 .. 1.0

    """

    _separable = False

    standard_broadcasting = False
    fittable = False

    n_inputs = 2
    n_outputs = 2

    crpix = Parameter(default=[0.0, 0.0], description="crpix")
    crval = Parameter(default=[0.0, 0.0], description="crval")
    cdelt = Parameter(default=[1.0, 1.0], description="cdelt")
    pc = Parameter(default=[[1.0, 0.0], [0.0, 1.0]], description="pc")

    def __init__(
        self, projection, crpix=crpix, crval=crval, cdelt=cdelt, pc=pc, **kwargs
    ):
        super().__init__(crpix=crpix, crval=crval, cdelt=cdelt, pc=pc, **kwargs)
        self.projection = projection
        self.inputs = ("x", "y")
        self.outputs = ("lon", "lat")
        self.lon_pole = _compute_lon_pole(self.crval, projection)

        self.forward = (
            Shift(-self.crpix[0]) & Shift(-self.crpix[1])
            | AffineTransformation2D(matrix=self.pc)
            | Scale(self.cdelt[0]) & Scale(self.cdelt[1])
            | self.projection
            | RotateNative2Celestial(self.crval[0], self.crval[1], self.lon_pole)
        )

    def evaluate(self, x, y, crpix, crval, cdelt, pc):
        return self.forward(x, y)

    def inverse(self):
        return self.forward.inverse



import numbers

import numpy as np
from astropy import units as u
from astropy.wcs.wcsapi.high_level_api import (
    high_level_objects_to_values,
    values_to_high_level_objects,
)

from ._base import BaseCoordinateFrame
from ._properties import FrameProperties

__all__ = ["CoordinateFrame"]


class CoordinateFrame(BaseCoordinateFrame):
    """
    Base class for Coordinate Frames.

    Parameters
    ----------
    naxes : int
        Number of axes.
    axes_type : str
        One of ["SPATIAL", "SPECTRAL", "TIME"]
    axes_order : tuple of int
        A dimension in the input data that corresponds to this axis.
    reference_frame : astropy.coordinates.builtin_frames
        Reference frame (usually used with output_frame to convert to world
        coordinate objects).
    unit : list of astropy.units.Unit
        Unit for each axis.
    axes_names : list
        Names of the axes in this frame.
    name : str
        Name of this frame.
    """

    def __init__(
        self,
        naxes,
        axes_type,
        axes_order,
        reference_frame=None,
        unit=None,
        axes_names=None,
        name=None,
        axis_physical_types=None,
    ):
        self._naxes = naxes
        self._axes_order = tuple(axes_order)
        self._reference_frame = reference_frame

        if name is None:
            self._name = self.__class__.__name__
        else:
            self._name = name

        if len(self._axes_order) != naxes:
            msg = "Length of axes_order does not match number of axes."
            raise ValueError(msg)

        if isinstance(axes_type, str):
            axes_type = (axes_type,)

        self._prop = FrameProperties(
            naxes,
            axes_type,
            unit,
            axes_names,
            axis_physical_types or self._default_axis_physical_types(axes_type),
        )

        super().__init__()

    def _default_axis_physical_types(self, axes_type):
        """
        The default physical types to use for this frame if none are specified
        by the user.
        """
        return tuple(f"custom:{t}" for t in axes_type)

    def __repr__(self):
        fmt = (
            f'<{self.__class__.__name__}(name="{self.name}", unit={self.unit}, '
            f"axes_names={self.axes_names}, axes_order={self.axes_order}"
        )
        if self.reference_frame is not None:
            fmt += f", reference_frame={self.reference_frame}"
        fmt += ")>"
        return fmt

    def __str__(self):
        if self._name is not None:
            return self._name
        return self.__class__.__name__

    def _sort_property(self, prop):
        sorted_prop = sorted(
            zip(prop, self.axes_order, strict=False), key=lambda x: x[1]
        )
        return tuple([t[0] for t in sorted_prop])

    @property
    def name(self):
        """A custom name of this frame."""
        return self._name

    @name.setter
    def name(self, val):
        """A custom name of this frame."""
        self._name = val

    @property
    def naxes(self):
        """The number of axes in this frame."""
        return self._naxes

    @property
    def unit(self):
        """The unit of this frame."""
        return self._sort_property(self._prop.unit)

    @property
    def axes_names(self):
        """Names of axes in the frame."""
        return self._sort_property(self._prop.axes_names)

    @property
    def axes_order(self):
        """A tuple of indices which map inputs to axes."""
        return self._axes_order

    @property
    def reference_frame(self):
        """Reference frame, used to convert to world coordinate objects."""
        return self._reference_frame

    @property
    def axes_type(self):
        """Type of this frame : 'SPATIAL', 'SPECTRAL', 'TIME'."""
        return self._sort_property(self._prop.axes_type)

    @property
    def axis_physical_types(self):
        """
        The axis physical types for this frame.

        These physical types are the types in frame order, not transform order.
        """
        return self._sort_property(self._prop.axis_physical_types)

    @property
    def world_axis_object_classes(self):
        return {
            f"{at}{i}" if i != 0 else at: (u.Quantity, (), {"unit": unit})
            for i, (at, unit) in enumerate(zip(self.axes_type, self.unit, strict=False))
        }

    @property
    def _native_world_axis_object_components(self):
        return [
            (f"{at}{i}" if i != 0 else at, 0, "value")
            for i, at in enumerate(self._prop.axes_type)
        ]

    @property
    def serialized_classes(self):
        """
        This property is used by the low level WCS API in Astropy.

        By providing it we can duck type as a low level WCS object.
        """
        return False

    def to_high_level_coordinates(self, *values):
        """
        Convert "values" to high level coordinate objects described by this frame.

        "values" are the coordinates in array or scalar form, and high level
        objects are things such as ``SkyCoord`` or ``Quantity``. See
        :ref:`wcsapi` for details.

        Parameters
        ----------
        values : `numbers.Number`, `numpy.ndarray`, or `~astropy.units.Quantity`
           ``naxis`` number of coordinates as scalars or arrays.

        Returns
        -------
        high_level_coordinates
            One (or more) high level object describing the coordinate.
        """
        # We allow Quantity-like objects here which values_to_high_level_objects
        # does not.
        values = [
            v.to_value(unit) if hasattr(v, "to_value") else v
            for v, unit in zip(values, self.unit, strict=False)
        ]

        if not all(
            isinstance(v, numbers.Number) or type(v) is np.ndarray for v in values
        ):
            msg = "All values should be a scalar number or a numpy array."
            raise TypeError(msg)

        high_level = values_to_high_level_objects(*values, low_level_wcs=self)
        if len(high_level) == 1:
            high_level = high_level[0]
        return high_level

    def from_high_level_coordinates(self, *high_level_coords):
        """
        Convert high level coordinate objects to "values" as described by this frame.

        "values" are the coordinates in array or scalar form, and high level
        objects are things such as ``SkyCoord`` or ``Quantity``. See
        :ref:`wcsapi` for details.

        Parameters
        ----------
        high_level_coordinates
            One (or more) high level object describing the coordinate.

        Returns
        -------
        values : `numbers.Number` or `numpy.ndarray`
           ``naxis`` number of coordinates as scalars or arrays.
        """
        values = high_level_objects_to_values(*high_level_coords, low_level_wcs=self)
        if len(values) == 1:
            values = values[0]
        return values



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import asdf
from asdf_astropy.testing.helpers import (
    assert_model_equal,
)
from astropy import coordinates as coord
from astropy import time
from astropy import units as u
from astropy.modeling import models

from gwcs import coordinate_frames as cf
from gwcs import wcs


def _assert_frame_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)

    if a is None:
        return None

    if not isinstance(a, cf.CoordinateFrame):
        return a == b

    assert a.name == b.name  # nosec
    if not isinstance(a, cf.EmptyFrame):
        assert a.axes_order == b.axes_order  # nosec
        assert a.axes_names == b.axes_names  # nosec
        assert a.unit == b.unit  # nosec
        assert a.reference_frame == b.reference_frame  # nosec
    return None


def assert_frame_roundtrip(frame, tmp_path, version=None):
    """
    Assert that a frame can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    path = tmp_path / "test.asdf"

    with asdf.AsdfFile({"frame": frame}, version=version) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        _assert_frame_equal(frame, af["frame"])


def _assert_wcs_equal(a, b):
    assert a.name == b.name  # nosec
    assert a.pixel_shape == b.pixel_shape
    assert len(a.available_frames) == len(b.available_frames)  # nosec
    for a_step, b_step in zip(a.pipeline, b.pipeline, strict=False):
        _assert_frame_equal(a_step.frame, b_step.frame)
        assert_model_equal(a_step.transform, b_step.transform)


def assert_wcs_roundtrip(wcs, tmp_path, version=None):
    path = tmp_path / "test.asdf"

    with asdf.AsdfFile({"wcs": wcs}, version=version) as af:
        af.write_to(path)

    with asdf.open(path) as af:
        _assert_wcs_equal(wcs, af["wcs"])


def test_create_wcs(tmp_path):
    m1 = models.Shift(12.4) & models.Shift(-2)
    icrs = cf.CelestialFrame(name="icrs", reference_frame=coord.ICRS())
    det = cf.Frame2D(name="detector", axes_order=(0, 1))
    gw1 = wcs.WCS(output_frame="icrs", input_frame="detector", forward_transform=m1)
    gw2 = wcs.WCS(output_frame="icrs", forward_transform=m1)
    gw3 = wcs.WCS(output_frame=icrs, input_frame=det, forward_transform=m1)
    gw4 = wcs.WCS(output_frame=icrs, input_frame=det, forward_transform=m1)
    gw4.pixel_shape = (100, 200)

    assert_wcs_roundtrip(gw1, tmp_path)
    assert_wcs_roundtrip(gw2, tmp_path)
    assert_wcs_roundtrip(gw3, tmp_path)
    assert_wcs_roundtrip(gw4, tmp_path)


def test_composite_frame(tmp_path):
    icrs = coord.ICRS()
    fk5 = coord.FK5()
    cel1 = cf.CelestialFrame(reference_frame=icrs)
    cel2 = cf.CelestialFrame(reference_frame=fk5)

    spec1 = cf.SpectralFrame(name="freq", unit=(u.Hz,), axes_order=(2,))
    spec2 = cf.SpectralFrame(name="wave", unit=(u.m,), axes_order=(2,))

    comp1 = cf.CompositeFrame([cel1, spec1])
    comp2 = cf.CompositeFrame([cel2, spec2])
    comp = cf.CompositeFrame([comp1, cf.SpectralFrame(axes_order=(3,), unit=(u.m,))])

    assert_frame_roundtrip(comp, tmp_path)
    assert_frame_roundtrip(comp1, tmp_path)
    assert_frame_roundtrip(comp2, tmp_path)


def create_test_frames():
    """Creates an array of frames to be used for testing."""

    return [
        cf.CelestialFrame(reference_frame=coord.ICRS()),
        cf.CelestialFrame(reference_frame=coord.FK5(equinox=time.Time("2010-01-01"))),
        cf.CelestialFrame(
            reference_frame=coord.FK4(
                equinox=time.Time("2010-01-01"), obstime=time.Time("2015-01-01")
            )
        ),
        cf.CelestialFrame(
            reference_frame=coord.FK4NoETerms(
                equinox=time.Time("2010-01-01"), obstime=time.Time("2015-01-01")
            )
        ),
        cf.CelestialFrame(reference_frame=coord.Galactic()),
        cf.CelestialFrame(
            reference_frame=coord.Galactocentric(
                # A default galcen_coord is used since none is provided here
                galcen_distance=5.0 * u.m,
                z_sun=3 * u.pc,
                roll=3 * u.deg,
            )
        ),
        cf.CelestialFrame(
            reference_frame=coord.GCRS(
                obstime=time.Time("2010-01-01"),
                obsgeoloc=[1, 3, 2000] * u.pc,
                obsgeovel=[2, 1, 8] * (u.m / u.s),
            )
        ),
        cf.CelestialFrame(reference_frame=coord.CIRS(obstime=time.Time("2010-01-01"))),
        cf.CelestialFrame(reference_frame=coord.ITRS(obstime=time.Time("2022-01-03"))),
        cf.CelestialFrame(
            reference_frame=coord.PrecessedGeocentric(
                obstime=time.Time("2010-01-01"),
                obsgeoloc=[1, 3, 2000] * u.pc,
                obsgeovel=[2, 1, 8] * (u.m / u.s),
            )
        ),
        cf.StokesFrame(),
        cf.TemporalFrame(time.Time("2011-01-01")),
    ]


def test_frames(tmp_path):
    frames = create_test_frames()
    for f in frames:
        assert_frame_roundtrip(f, tmp_path)


def test_references(tmp_path):
    m1 = models.Shift(12.4) & models.Shift(-2)
    icrs = cf.CelestialFrame(name="icrs", reference_frame=coord.ICRS())
    det = cf.Frame2D(name="detector", axes_order=(0, 1))
    det2 = cf.Frame2D(name="detector2", axes_order=(0, 1))
    focal = cf.Frame2D(name="focal", axes_order=(0, 1))

    pipe1 = [(det, m1), (focal, m1), (icrs, None)]
    gw1 = wcs.WCS(pipe1)

    pipe2 = [(det, m1), (det2, m1), (icrs, None)]
    gw2 = wcs.WCS(pipe2)

    tree = {"wcs1": gw1, "wcs2": gw2}
    af = asdf.AsdfFile(tree)
    output_path = tmp_path / "test.asdf"
    af.write_to(output_path)

    with asdf.open(output_path) as af:
        gw1 = af.tree["wcs1"]
        gw2 = af.tree["wcs2"]
        assert gw1.pipeline[0].transform is gw1.pipeline[1].transform
        assert gw2.pipeline[0].transform is gw2.pipeline[1].transform
        assert gw2.pipeline[0].frame is gw1.pipeline[0].frame



from dataclasses import InitVar, dataclass

import numpy as np
from astropy import units as u
from astropy.wcs.wcsapi.low_level_api import VALID_UCDS, validate_physical_types

__all__ = ["FrameProperties"]


@dataclass
class FrameProperties:
    naxes: InitVar[int]
    axes_type: tuple[str, ...]
    unit: tuple[u.Unit, ...] | None = None
    axes_names: tuple[str, ...] | None = None
    axis_physical_types: tuple[str, ...] | None = None

    def __post_init__(self, naxes: int):
        if isinstance(self.axes_type, str):
            self.axes_type = (self.axes_type,)
        else:
            self.axes_type = tuple(self.axes_type)

        if len(self.axes_type) != naxes:
            msg = "Length of axes_type does not match number of axes."
            raise ValueError(msg)

        if self.unit is not None:
            unit = tuple(self.unit) if np.iterable(self.unit) else (self.unit,)
            if len(unit) != naxes:
                msg = "Number of units does not match number of axes."
                raise ValueError(msg)
            self.unit = tuple(u.Unit(au) for au in unit)
        else:
            self.unit = tuple(u.dimensionless_unscaled for na in range(naxes))

        if self.axes_names is not None:
            if isinstance(self.axes_names, str):
                self.axes_names = (self.axes_names,)
            else:
                self.axes_names = tuple(self.axes_names)
            if len(self.axes_names) != naxes:
                msg = "Number of axes names does not match number of axes."
                raise ValueError(msg)
        else:
            self.axes_names = tuple([""] * naxes)

        if self.axis_physical_types is not None:
            if isinstance(self.axis_physical_types, str):
                self.axis_physical_types = (self.axis_physical_types,)
            elif not np.iterable(self.axis_physical_types):
                msg = (
                    "axis_physical_types must be of type string or iterable of strings"
                )
                raise TypeError(msg)
            if len(self.axis_physical_types) != naxes:
                msg = f'"axis_physical_types" must be of length {naxes}'
                raise ValueError(msg)
            ph_type = []
            for axt in self.axis_physical_types:
                if axt not in VALID_UCDS and not axt.startswith("custom:"):
                    ph_type.append(f"custom:{axt}")
                else:
                    ph_type.append(axt)

            validate_physical_types(ph_type)
            self.axis_physical_types = tuple(ph_type)

    @property
    def _default_axis_physical_types(self):
        """
        The default physical types to use for this frame if none are specified
        by the user.
        """
        return tuple(f"custom:{t}" for t in self.axes_type)



import abc

import numpy as np
from astropy import units as u

from ._properties import FrameProperties

__all__ = ["BaseCoordinateFrame"]


class BaseCoordinateFrame(abc.ABC):
    """
    API Definition for a Coordinate frame
    """

    _prop: FrameProperties
    """
    The FrameProperties object holding properties in native frame order.
    """

    @property
    @abc.abstractmethod
    def naxes(self) -> int:
        """
        The number of axes described by this frame.
        """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        The name of the coordinate frame.
        """

    @property
    @abc.abstractmethod
    def unit(self) -> tuple[u.Unit, ...]:
        """
        The units of the axes in this frame.
        """

    @property
    @abc.abstractmethod
    def axes_names(self) -> tuple[str, ...]:
        """
        Names describing the axes of the frame.
        """

    @property
    @abc.abstractmethod
    def axes_order(self) -> tuple[int, ...]:
        """
        The position of the axes in the frame in the transform.
        """

    @property
    @abc.abstractmethod
    def reference_frame(self):
        """
        The reference frame of the coordinates described by this frame.

        This is usually an Astropy object such as ``SkyCoord`` or ``Time``.
        """

    @property
    @abc.abstractmethod
    def axes_type(self):
        """
        An upcase string describing the type of the axis.

        Known values are ``"SPATIAL", "TEMPORAL", "STOKES", "SPECTRAL", "PIXEL"``.
        """

    @property
    @abc.abstractmethod
    def axis_physical_types(self):
        """
        The UCD 1+ physical types for the axes, in frame order.
        """

    @property
    @abc.abstractmethod
    def world_axis_object_classes(self):
        """
        The APE 14 object classes for this frame.

        See Also
        --------
        astropy.wcs.wcsapi.BaseLowLevelWCS.world_axis_object_classes
        """

    @property
    def world_axis_object_components(self):
        """
        The APE 14 object components for this frame.

        See Also
        --------
        astropy.wcs.wcsapi.BaseLowLevelWCS.world_axis_object_components
        """
        if self.naxes == 1:
            return self._native_world_axis_object_components

        # If we have more than one axis then we should sort the native
        # components by the axes_order.
        ordered = np.array(self._native_world_axis_object_components, dtype=object)[
            np.argsort(self.axes_order)
        ]
        return list(map(tuple, ordered))

    @property
    @abc.abstractmethod
    def _native_world_axis_object_components(self):
        """
        This property holds the "native" frame order of the components.

        The native order of the components is the order the frame assumes the
        axes are in when creating the high level objects, for example
        ``CelestialFrame`` creates ``SkyCoord`` objects which are in lon, lat
        order (in their positional args).

        This property is used both to construct the ordered
        ``world_axis_object_components`` property as well as by `CompositeFrame`
        to be able to get the components in their native order.
        """

    def add_units(
        self, arrays: u.Quantity | np.ndarray | list[float]
    ) -> tuple[u.Quantity, ...] | u.Quantity:
        """
        Add units to the arrays
        """
        if self.naxes == 1 and np.isscalar(arrays):
            return u.Quantity(arrays, self.unit[0])

        return tuple(
            u.Quantity(array, unit=unit)
            for array, unit in zip(arrays, self.unit, strict=True)
        )

    def remove_units(
        self, arrays: u.Quantity | np.ndarray | list[float]
    ) -> tuple[np.ndarray, ...]:
        """
        Remove units from the input arrays
        """
        if self.naxes == 1 and (np.isscalar(arrays) or isinstance(arrays, u.Quantity)):
            arrays = (arrays,)

        return tuple(
            array.to_value(unit) if isinstance(array, u.Quantity) else array
            for array, unit in zip(arrays, self.unit, strict=True)
        )



import astropy.units as u
import numpy as np
from astropy import coordinates as coord
from astropy.modeling import models
from astropy.time import Time
from astropy.wcs import WCS as ASTWCS

from . import coordinate_frames as cf
from . import fitswcs, geometry, wcs
from . import spectroscopy as sp

# frames
DETECTOR_1D_FRAME = cf.CoordinateFrame(
    name="detector", axes_order=(0,), naxes=1, axes_type="detector"
)
DETECTOR_2D_FRAME = cf.Frame2D(name="detector", axes_order=(0, 1))
ICRC_SKY_FRAME = cf.CelestialFrame(reference_frame=coord.ICRS(), axes_order=(0, 1))

FREQ_FRAME = cf.SpectralFrame(name="freq", unit=u.Hz, axes_order=(0,))
WAVE_FRAME = cf.SpectralFrame(
    name="wave", unit=u.m, axes_order=(2,), axes_names=("lambda",)
)

# transforms
MODEL_2D_SHIFT = models.Shift(1) & models.Shift(2)
MODEL_1D_SCALE = models.Scale(2)


def gwcs_simple_2d():
    output_frame = cf.Frame2D(name="world")
    return wcs.WCS(MODEL_2D_SHIFT, output_frame=output_frame)


def gwcs_empty_output_2d():
    return wcs.WCS(MODEL_2D_SHIFT, output_frame=cf.EmptyFrame(name="world"))


def gwcs_2d_quantity_shift():
    frame = cf.CoordinateFrame(
        name="quantity",
        axes_order=(0, 1),
        naxes=2,
        axes_type=("SPATIAL", "SPATIAL"),
        unit=(u.km, u.km),
    )
    pipe = [(DETECTOR_2D_FRAME, MODEL_2D_SHIFT), (frame, None)]

    return wcs.WCS(pipe)


def gwcs_2d_spatial_shift():
    """
    A simple one step spatial WCS, in ICRS with a 1 and 2 px shift.
    """
    pipe = [(DETECTOR_2D_FRAME, MODEL_2D_SHIFT), (ICRC_SKY_FRAME, None)]
    return wcs.WCS(pipe)


def gwcs_2d_spatial_reordered():
    """
    A simple one step spatial WCS, in ICRS with a 1 and 2 px shift.
    """
    out_frame = cf.CelestialFrame(reference_frame=coord.ICRS(), axes_order=(1, 0))
    return wcs.WCS(
        MODEL_2D_SHIFT | models.Mapping((1, 0)),
        input_frame=DETECTOR_2D_FRAME,
        output_frame=out_frame,
    )


def gwcs_1d_freq():
    return wcs.WCS([(DETECTOR_1D_FRAME, MODEL_1D_SCALE), (FREQ_FRAME, None)])


def gwcs_3d_spatial_wave():
    comp1 = cf.CompositeFrame([ICRC_SKY_FRAME, WAVE_FRAME])
    m = MODEL_2D_SHIFT & MODEL_1D_SCALE

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=3,
        axes_order=(0, 1, 2),
        axes_type=("pixel", "pixel", "pixel"),
        axes_names=("x", "y", "z"),
        unit=(u.pix, u.pix, u.pix),
    )

    return wcs.WCS([(detector_frame, m), (comp1, None)])


def gwcs_2d_shift_scale():
    m1 = models.Shift(1) & models.Shift(2)
    m2 = models.Scale(5) & models.Scale(10)
    m3 = m1 | m2
    pipe = [(DETECTOR_2D_FRAME, m3), (ICRC_SKY_FRAME, None)]
    return wcs.WCS(pipe)


def gwcs_2d_bad_bounding_box_order():
    m1 = models.Shift(1) & models.Shift(2)
    m2 = models.Scale(5) & models.Scale(10)
    m3 = m1 | m2

    # Purposefully set the bounding box in the wrong order
    m3.bounding_box = ((1, 2), (3, 4))

    pipe = [(DETECTOR_2D_FRAME, m3), (ICRC_SKY_FRAME, None)]
    return wcs.WCS(pipe)


def gwcs_1d_freq_quantity():
    detector_1d = cf.CoordinateFrame(
        name="detector", axes_order=(0,), naxes=1, unit=u.pix, axes_type="detector"
    )
    return wcs.WCS(
        [(detector_1d, models.Multiply(1 * u.Hz / u.pix)), (FREQ_FRAME, None)]
    )


def gwcs_2d_shift_scale_quantity():
    m4 = models.Shift(1 * u.pix) & models.Shift(2 * u.pix)
    m5 = models.Scale(5 * u.deg)
    m6 = models.Scale(10 * u.deg)
    m5.input_units_equivalencies = {"x": u.pixel_scale(1 * u.deg / u.pix)}
    m6.input_units_equivalencies = {"x": u.pixel_scale(1 * u.deg / u.pix)}
    m5.inverse = models.Scale(1.0 / 5 * u.pix)
    m6.inverse = models.Scale(1.0 / 10 * u.pix)
    m5.inverse.input_units_equivalencies = {"x": u.pixel_scale(1 * u.pix / u.deg)}
    m6.inverse.input_units_equivalencies = {"x": u.pixel_scale(1 * u.pix / u.deg)}
    m7 = m5 & m6
    m8 = m4 | m7
    pipe2 = [(DETECTOR_2D_FRAME, m8), (ICRC_SKY_FRAME, None)]
    return wcs.WCS(pipe2)


def gwcs_3d_identity_units():
    """
    A simple 1-1 gwcs that converts from pixels to arcseconds
    """
    identity = (
        models.Multiply(1 * u.arcsec / u.pixel)
        & models.Multiply(1 * u.arcsec / u.pixel)
        & models.Multiply(1 * u.nm / u.pixel)
    )
    sky_frame = cf.CelestialFrame(
        axes_order=(0, 1),
        name="icrs",
        reference_frame=coord.ICRS(),
        axes_names=("longitude", "latitude"),
        unit=(u.arcsec, u.arcsec),
    )
    wave_frame = cf.SpectralFrame(
        axes_order=(2,), unit=u.nm, axes_names=("wavelength",)
    )

    frame = cf.CompositeFrame([sky_frame, wave_frame])

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=3,
        axes_order=(0, 1, 2),
        axes_type=("pixel", "pixel", "pixel"),
        axes_names=("x", "y", "z"),
        unit=(u.pix, u.pix, u.pix),
    )

    return wcs.WCS(
        forward_transform=identity, output_frame=frame, input_frame=detector_frame
    )


def gwcs_4d_identity_units():
    """
    A simple 1-1 gwcs that converts from pixels to arcseconds
    """
    identity = (
        models.Multiply(1 * u.arcsec / u.pixel)
        & models.Multiply(1 * u.arcsec / u.pixel)
        & models.Multiply(1 * u.nm / u.pixel)
        & models.Multiply(1 * u.s / u.pixel)
    )
    sky_frame = cf.CelestialFrame(
        axes_order=(0, 1), name="icrs", reference_frame=coord.ICRS()
    )
    wave_frame = cf.SpectralFrame(axes_order=(2,), unit=u.nm)
    time_frame = cf.TemporalFrame(
        axes_order=(3,), unit=u.s, reference_frame=Time("2000-01-01T00:00:00")
    )

    frame = cf.CompositeFrame([sky_frame, wave_frame, time_frame])

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=4,
        axes_order=(0, 1, 2, 3),
        axes_type=("pixel", "pixel", "pixel", "pixel"),
        axes_names=("x", "y", "z", "s"),
        unit=(u.pix, u.pix, u.pix, u.pix),
    )

    return wcs.WCS(
        forward_transform=identity, output_frame=frame, input_frame=detector_frame
    )


def gwcs_simple_imaging_units():
    shift_by_crpix = models.Shift(-2048 * u.pix) & models.Shift(-1024 * u.pix)
    matrix = np.array(
        [
            [1.290551569736e-05, 5.9525007864732e-06],
            [5.0226382102765e-06, -1.2644844123757e-05],
        ]
    )
    rotation = models.AffineTransformation2D(matrix * u.deg, translation=[0, 0] * u.deg)
    rotation.input_units_equivalencies = {
        "x": u.pixel_scale(1 * u.deg / u.pix),
        "y": u.pixel_scale(1 * u.deg / u.pix),
    }
    rotation.inverse = models.AffineTransformation2D(
        np.linalg.inv(matrix) * u.pix, translation=[0, 0] * u.pix
    )
    rotation.inverse.input_units_equivalencies = {
        "x": u.pixel_scale(1 * u.pix / u.deg),
        "y": u.pixel_scale(1 * u.pix / u.deg),
    }
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        5.63056810618 * u.deg, -72.05457184279 * u.deg, 180 * u.deg
    )
    det2sky = shift_by_crpix | rotation | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    return wcs.WCS(pipeline)


def gwcs_simple_imaging():
    shift_by_crpix = models.Shift(-2048) & models.Shift(-1024)
    matrix = np.array(
        [
            [1.290551569736e-05, 5.9525007864732e-06],
            [5.0226382102765e-06, -1.2644844123757e-05],
        ]
    )
    rotation = models.AffineTransformation2D(matrix, translation=[0, 0])
    rotation.inverse = models.AffineTransformation2D(
        np.linalg.inv(matrix), translation=[0, 0]
    )
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        5.63056810618, -72.05457184279, 180
    )
    det2sky = shift_by_crpix | rotation | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(name="detector", axes_names=("x", "y"))
    sky_frame = cf.CelestialFrame(reference_frame=coord.ICRS(), name="icrs")
    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    return wcs.WCS(pipeline)


def gwcs_stokes_lookup():
    transform = models.Tabular1D(
        [0, 1, 2, 3] * u.pix,
        [1, 2, 3, 4] * u.one,
        method="nearest",
        fill_value=np.nan,
        bounds_error=False,
    )
    frame = cf.StokesFrame()

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=1,
        axes_order=(0,),
        axes_type=("pixel",),
        axes_names=("x",),
        unit=(u.pix,),
    )

    return wcs.WCS(
        forward_transform=transform, output_frame=frame, input_frame=detector_frame
    )


def gwcs_3spectral_orders():
    comp1 = cf.CompositeFrame([ICRC_SKY_FRAME, WAVE_FRAME])
    detector_frame = cf.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    m = models.Mapping((0, 1, 1)) | MODEL_2D_SHIFT & MODEL_1D_SCALE

    return wcs.WCS([(detector_frame, m), (comp1, None)])


def gwcs_with_frames_strings():
    transform = models.Shift(1) & models.Shift(1) & models.Polynomial2D(1)
    pipe = [("detector", transform), ("world", None)]
    return wcs.WCS(pipe)


def sellmeier_glass():
    B_coef = [0.58339748, 0.46085267, 3.8915394]
    C_coef = [0.00252643, 0.010078333, 1200.556]
    return sp.SellmeierGlass(B_coef, C_coef)


def sellmeier_zemax():
    B_coef = [0.58339748, 0.46085267, 3.8915394]
    C_coef = [0.00252643, 0.010078333, 1200.556]
    D_coef = [-2.66e-05, 0.0, 0.0]
    E_coef = [0.0, 0.0, 0.0]
    return sp.SellmeierZemax(
        65, 35, 0, 0, B_coef=B_coef, C_coef=C_coef, D_coef=D_coef, E_coef=E_coef
    )


def gwcs_3d_galactic_spectral():
    """
    This fixture has the axes ordered as lat, spectral, lon.
    """
    #                       lat,wav,lon
    crpix1, crpix2, crpix3 = 29, 39, 44
    crval1, crval2, crval3 = 10, 20, 25
    cdelt1, cdelt2, cdelt3 = -0.01, 0.5, 0.01

    shift = models.Shift(-crpix3) & models.Shift(-crpix1)
    scale = models.Multiply(cdelt3) & models.Multiply(cdelt1)
    proj = models.Pix2Sky_TAN()
    skyrot = models.RotateNative2Celestial(crval3, 90 + crval1, 180)
    celestial = shift | scale | proj | skyrot

    wave_model = models.Shift(-crpix2) | models.Multiply(cdelt2) | models.Shift(crval2)

    transform = (
        models.Mapping((2, 0, 1)) | celestial & wave_model | models.Mapping((1, 2, 0))
    )

    sky_frame = cf.CelestialFrame(
        axes_order=(2, 0),
        reference_frame=coord.Galactic(),
        axes_names=("Longitude", "Latitude"),
    )
    wave_frame = cf.SpectralFrame(axes_order=(1,), unit=u.Hz, axes_names=("Frequency",))

    frame = cf.CompositeFrame([sky_frame, wave_frame])

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=3,
        axes_order=(0, 1, 2),
        axes_type=("pixel", "pixel", "pixel"),
        unit=(u.pix, u.pix, u.pix),
    )

    owcs = wcs.WCS(
        forward_transform=transform, output_frame=frame, input_frame=detector_frame
    )
    owcs.bounding_box = ((-1, 35), (-2, 45), (5, 50))
    owcs.array_shape = (30, 20, 10)
    owcs.pixel_shape = (10, 20, 30)

    return owcs


def gwcs_1d_spectral():
    """
    A simple 1D spectral WCS.
    """
    wave_model = models.Shift(-5) | models.Multiply(3.7) | models.Shift(20)
    wave_model.bounding_box = (7, 50)
    wave_frame = cf.SpectralFrame(axes_order=(0,), unit=u.Hz, axes_names=("Frequency",))

    detector_frame = cf.CoordinateFrame(
        name="detector", naxes=1, axes_order=(0,), axes_type=("pixel",), unit=(u.pix,)
    )

    owcs = wcs.WCS(
        forward_transform=wave_model,
        output_frame=wave_frame,
        input_frame=detector_frame,
    )
    owcs.array_shape = (44,)
    owcs.pixel_shape = (44,)

    return owcs


def gwcs_spec_cel_time_4d():
    """
    A complex 4D mixed celestial + spectral + time WCS.
    """
    # spectroscopic frame:
    wave_model = models.Shift(-5) | models.Multiply(3.7) | models.Shift(20)
    wave_model.bounding_box = (7, 50)
    wave_frame = cf.SpectralFrame(
        name="wave", unit=u.m, axes_order=(0,), axes_names=("lambda",)
    )

    # time frame:
    time_model = models.Identity(1)  # models.Linear1D(10, 0)
    time_frame = cf.TemporalFrame(
        Time("2010-01-01T00:00"), name="time", unit=u.s, axes_order=(3,)
    )

    # Values from data/acs.hdr:
    crpix = (12, 13)
    crval = (5.63, -72.05)
    cd = [[1.291e-05, 5.9532e-06], [5.02215e-06, -1.2645e-05]]
    aff = models.AffineTransformation2D(matrix=cd, name="rotation")
    offx = models.Shift(-crpix[0], name="x_translation")
    offy = models.Shift(-crpix[1], name="y_translation")
    wcslin = models.Mapping((1, 0)) | (offx & offy) | aff
    tan = models.Pix2Sky_TAN(name="tangent_projection")
    n2c = models.RotateNative2Celestial(*crval, 180, name="sky_rotation")
    cel_model = wcslin | tan | n2c | models.Mapping((1, 0))
    icrs = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="sky", axes_order=(2, 1)
    )

    wcs_forward = wave_model & cel_model & time_model

    comp_frm = cf.CompositeFrame(
        frames=[wave_frame, icrs, time_frame], name="TEST 4D FRAME"
    )

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=4,
        axes_order=(0, 1, 2, 3),
        axes_type=("pixel", "pixel", "pixel", "pixel"),
        unit=(u.pix, u.pix, u.pix, u.pix),
    )

    w = wcs.WCS(
        forward_transform=wcs_forward, output_frame=comp_frm, input_frame=detector_frame
    )

    w.bounding_box = ((0, 63), (0, 127), (0, 255), (0, 9))
    w.array_shape = (10, 256, 128, 64)
    w.pixel_shape = (64, 128, 256, 10)
    return w


def gwcs_cube_with_separable_spectral(axes_order):
    """
    GWCS cube with spectral axis separable from the celestial axes.

    Viable examples are (2, 0, 1) and (2, 1, 0).
        (1, 0, 2) fails round-trip for -TAB axis 3
    """
    cube_size = (128, 64, 100)

    spectral_axes_order = (axes_order.index(2),)
    cel_axes_order = (axes_order.index(0), axes_order.index(1))

    # Values from data/acs.hdr:
    crpix = (64, 32)
    crval = (5.63056810618, -72.0545718428)
    cd = [
        [1.29058667557984e-05, 5.95320245884555e-06],
        [5.02215195623825e-06, -1.2645010396976e-05],
    ]

    aff = models.AffineTransformation2D(matrix=cd, name="rotation")
    offx = models.Shift(-crpix[0], name="x_translation")
    offy = models.Shift(-crpix[1], name="y_translation")

    wcslin = (offx & offy) | aff
    tan = models.Pix2Sky_TAN(name="tangent_projection")
    n2c = models.RotateNative2Celestial(*crval, 180, name="sky_rotation")
    icrs = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="sky", axes_order=cel_axes_order
    )
    spec = cf.SpectralFrame(
        name="wave",
        unit=[
            u.m,
        ],
        axes_order=spectral_axes_order,
        axes_names=("lambda",),
    )
    comp_frm = cf.CompositeFrame(
        frames=[icrs, spec], name="TEST 3D FRAME WITH SPECTRAL AXIS"
    )
    wcs_forward = (
        (wcslin & models.Identity(1))
        | (tan & models.Identity(1))
        | (n2c & models.Identity(1))
        | models.Mapping(axes_order)
    )

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=3,
        axes_order=(0, 1, 2),
        axes_type=("pixel", "pixel", "pixel"),
        unit=(u.pix, u.pix, u.pix),
    )

    w = wcs.WCS(
        forward_transform=wcs_forward, output_frame=comp_frm, input_frame=detector_frame
    )
    w.bounding_box = tuple((0, k - 1) for k in cube_size)
    w.pixel_shape = cube_size
    w.array_shape = w.pixel_shape[::-1]

    return w, axes_order


def gwcs_cube_with_separable_time(axes_order):
    """
    A mixed celestial + time WCS.

    Viable examples are (2, 0, 1) and (2, 1, 0).
        (0, 2, 1) fails round-trip for -TAB axis 2
        (1, 0, 2) fails round-trip for -TAB axis 3
    """
    cube_size = (64, 32, 128)

    time_axes_order = (axes_order.index(2),)
    cel_axes_order = (axes_order.index(0), axes_order.index(1))

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=3,
        axes_order=(0, 1, 2),
        axes_type=("pixel", "pixel", "pixel"),
        unit=(u.pix, u.pix, u.pix),
    )

    # time frame:
    time_model = models.Identity(1)  # models.Linear1D(10, 0)
    time_frame = cf.TemporalFrame(
        Time("2010-01-01T00:00"), name="time", unit=u.s, axes_order=time_axes_order
    )

    # Values from data/acs.hdr:
    crpix = (12, 13)
    crval = (5.63, -72.05)
    cd = [[1.291e-05, 5.9532e-06], [5.02215e-06, -1.2645e-05]]
    aff = models.AffineTransformation2D(matrix=cd, name="rotation")
    offx = models.Shift(-crpix[0], name="x_translation")
    offy = models.Shift(-crpix[1], name="y_translation")
    wcslin = models.Mapping((1, 0)) | (offx & offy) | aff
    tan = models.Pix2Sky_TAN(name="tangent_projection")
    n2c = models.RotateNative2Celestial(*crval, 180, name="sky_rotation")
    cel_model = wcslin | tan | n2c
    icrs = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="sky", axes_order=cel_axes_order
    )

    wcs_forward = (cel_model & time_model) | models.Mapping(axes_order)

    comp_frm = cf.CompositeFrame(
        frames=[icrs, time_frame], name="TEST 3D FRAME WITH TIME"
    )

    w = wcs.WCS(
        forward_transform=wcs_forward, output_frame=comp_frm, input_frame=detector_frame
    )

    w.bounding_box = tuple((0, k - 1) for k in cube_size)
    w.pixel_shape = cube_size
    w.array_shape = w.pixel_shape[::-1]
    return w


def gwcs_7d_complex_mapping():
    """
    Useful features of this WCS (axes indices here are 0-based):
        - includes two celestial axes: input (0, 1) maps to world (2 - RA, 1 - Dec)
        - includes one separable frame with one axis: 4 -> 2
        - includes one frame with 3 input and 4 output axes (1 degenerate),
          with separable world axes (3, 5) and (0, 6).
    """
    offx = models.Shift(-64, name="x_translation")
    offy = models.Shift(-32, name="y_translation")
    cd = np.array([[1.2906, 0.59532], [0.50222, -1.2645]])
    aff = models.AffineTransformation2D(matrix=1e-5 * cd, name="rotation")
    aff2 = models.AffineTransformation2D(matrix=cd, name="rotation2")

    wcslin = (offx & offy) | aff
    tan = models.Pix2Sky_TAN(name="tangent_projection")
    n2c = models.RotateNative2Celestial(5.630568, -72.0546, 180, name="skyrot")
    icrs = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="sky", axes_order=(2, 1)
    )
    spec = cf.SpectralFrame(
        name="wave", unit=[u.m], axes_order=(4,), axes_names=("lambda",)
    )
    cmplx = cf.CoordinateFrame(
        name="complex",
        naxes=4,
        axes_order=(3, 5, 0, 6),
        axis_physical_types=(["em.wl", "em.wl", "time", "time"]),
        axes_type=("SPATIAL", "SPATIAL", "TIME", "TIME"),
        axes_names=("x", "y", "t", "tau"),
        unit=(u.m, u.m, u.second, u.second),
    )

    comp_frm = cf.CompositeFrame(frames=[icrs, spec, cmplx], name="TEST 7D")
    wcs_forward = (
        (wcslin & models.Shift(-3.14) & models.Scale(2.7) & aff2)
        | (tan & models.Identity(1) & models.Identity(1) & models.Identity(2))
        | (n2c & models.Identity(1) & models.Identity(1) & models.Identity(2))
        | models.Mapping((3, 1, 0, 4, 2, 5, 3))
    )

    detector_frame = cf.CoordinateFrame(
        name="detector",
        naxes=6,
        axes_order=(0, 1, 2, 3, 4, 5),
        axes_type=("pixel", "pixel", "pixel", "pixel", "pixel", "pixel"),
        unit=(u.pix, u.pix, u.pix, u.pix, u.pix, u.pix),
    )

    w = wcs.WCS(
        forward_transform=wcs_forward, output_frame=comp_frm, input_frame=detector_frame
    )
    w.bounding_box = ((0, 15), (0, 31), (0, 20), (0, 10), (0, 10), (0, 1))

    w.array_shape = (2, 11, 11, 21, 32, 16)
    w.pixel_shape = (16, 32, 21, 11, 11, 2)

    return w


def gwcs_with_pipeline_celestial():
    input_frame = cf.CoordinateFrame(
        2, ["PIXEL"] * 2, axes_order=list(range(2)), unit=[u.pix] * 2, name="input"
    )

    spatial = models.Multiply(20 * u.arcsec / u.pix) & models.Multiply(
        15 * u.deg / u.pix
    )

    celestial_frame = cf.CelestialFrame(
        axes_order=(0, 1),
        unit=(u.arcsec, u.deg),
        reference_frame=coord.ICRS(),
        name="celestial",
    )

    custom = models.Shift(1 * u.deg) & models.Shift(2 * u.deg)

    output_frame = cf.CoordinateFrame(
        2, ["CUSTOM"] * 2, axes_order=list(range(2)), unit=[u.arcsec] * 2, name="output"
    )

    pipeline = [
        (input_frame, spatial),
        (celestial_frame, custom),
        (output_frame, None),
    ]

    return wcs.WCS(pipeline)


def gwcs_romanisim():
    targ_pos = coord.SkyCoord(ra=0 * u.deg, dec=0 * u.deg)

    ra_ref = targ_pos.ra.to(u.deg).value
    dec_ref = targ_pos.dec.to(u.deg).value

    # v2_ref, v3_ref are in arcsec, but RotationSequence3D wants degrees,
    # so start by scaling by 3600.
    rot = models.RotationSequence3D([0, 0, 0, dec_ref, -ra_ref], "zyxyz")

    # V2V3 are in arcseconds, while SphericalToCartesian expects degrees,
    # so again start by scaling by 3600
    tel2sky = (
        (models.Scale(1 / 3600) & models.Scale(1 / 3600))
        | geometry.SphericalToCartesian(wrap_lon_at=180)
        | rot
        | geometry.CartesianToSpherical(wrap_lon_at=360)
    )
    tel2sky.name = "v23tosky"

    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    v2v3vacorr = cf.Frame2D(
        name="v2v3vacorr",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    va_corr = models.Identity(2)
    zen2v2v3 = models.EulerAngleRotation(0, -90, 0, "xyz") | (
        models.Scale(3600) & models.Scale(3600)
    )
    tanproj = models.Pix2Sky_Gnomonic()
    pix2tan = models.Scale(0.11 / 3600) & models.Scale(0.11 / 3600)
    distortion = pix2tan | tanproj | zen2v2v3

    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, va_corr),
        wcs.Step(v2v3vacorr, tel2sky),
        wcs.Step(world, None),
    ]

    return wcs.WCS(pipeline)


def fits_wcs_imaging_simple(params):
    """A simple FITS WCS imaging transform without distortion."""
    # Check for several values of cravl2, including at the poles
    lon, lat = params
    # First, generate a GWCS object
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    world = cf.CelestialFrame(
        reference_frame=coord.ICRS(), name="world", unit=(u.deg, u.deg)
    )
    projection = models.Pix2Sky_TAN()
    crval = [lon, lat]
    crpix = [5, 5]
    fwcs = fitswcs.FITSImagingWCSTransform(projection, crpix=crpix, crval=crval)
    pipeline = [wcs.Step(detector, fwcs), wcs.Step(world, None)]
    gw = wcs.WCS(pipeline)

    # Generate a FITSWCS object
    w = ASTWCS(naxis=2)
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    # Add 1 to CRPIX for FITS compatibility
    w.wcs.crpix = np.array(crpix) + 1
    w.wcs.crval = crval
    if crval[1] in (90, -90):
        # following the errata WCS paper
        w.wcs.lonpole = 180
    w.wcs.set()
    return gw, w


def gwcs_2d_spatial_shift_reverse():
    """
    A simple one step spatial WCS with forward from sky to detector.
    """
    pipe = [(ICRC_SKY_FRAME, MODEL_2D_SHIFT), (DETECTOR_2D_FRAME, None)]
    return wcs.WCS(pipe)


def gwcs_multi_stage():
    """
    A 3-step pipeline where the intermediate step is 1D and the final is 2D.
    """
    tr1 = models.Shift(10)
    tr2 = models.Mapping((0, 0)) | models.Scale(-2) & models.Scale(-1)
    det = cf.CoordinateFrame(
        name="detector", naxes=1, unit=("pix",), axes_type="SPATIAL", axes_order=(0,)
    )
    intermediate = cf.CoordinateFrame(
        name="intermediate", naxes=1, unit=("m",), axes_type="SPATIAL", axes_order=(0,)
    )
    cel = cf.CelestialFrame(name="sky", axes_names=("ra", "dec"))
    return wcs.WCS([(det, tr1), (intermediate, tr2), (cel, None)])



"""
This file contains a set of pytest fixtures which are different gwcses for testing.
"""

import pytest

from gwcs import examples, geometry


@pytest.fixture
def gwcs_simple_2d():
    return examples.gwcs_simple_2d()


@pytest.fixture
def gwcs_empty_output_2d():
    return examples.gwcs_empty_output_2d()


@pytest.fixture
def gwcs_2d_quantity_shift():
    return examples.gwcs_2d_quantity_shift()


@pytest.fixture
def gwcs_2d_spatial_shift():
    return examples.gwcs_2d_spatial_shift()


@pytest.fixture
def gwcs_2d_spatial_reordered():
    return examples.gwcs_2d_spatial_reordered()


@pytest.fixture
def gwcs_1d_freq():
    return examples.gwcs_1d_freq()


@pytest.fixture
def gwcs_3d_spatial_wave():
    return examples.gwcs_3d_spatial_wave()


@pytest.fixture
def gwcs_2d_shift_scale():
    return examples.gwcs_2d_shift_scale()


@pytest.fixture
def gwcs_1d_freq_quantity():
    return examples.gwcs_1d_freq_quantity()


@pytest.fixture
def gwcs_2d_shift_scale_quantity():
    return examples.gwcs_2d_shift_scale_quantity()


@pytest.fixture
def gwcs_3d_identity_units():
    return examples.gwcs_3d_identity_units()


@pytest.fixture
def gwcs_4d_identity_units():
    return examples.gwcs_4d_identity_units()


@pytest.fixture
def gwcs_simple_imaging_units():
    return examples.gwcs_simple_imaging_units()


@pytest.fixture
def gwcs_simple_imaging():
    return examples.gwcs_simple_imaging()


@pytest.fixture
def gwcs_stokes_lookup():
    return examples.gwcs_stokes_lookup()


@pytest.fixture
def gwcs_3spectral_orders():
    return examples.gwcs_3spectral_orders()


@pytest.fixture
def gwcs_with_frames_strings():
    return examples.gwcs_with_frames_strings()


@pytest.fixture
def sellmeier_glass():
    return examples.sellmeier_glass()


@pytest.fixture
def sellmeier_zemax():
    return examples.sellmeier_zemax()


@pytest.fixture
def gwcs_3d_galactic_spectral():
    return examples.gwcs_3d_galactic_spectral()


@pytest.fixture
def gwcs_1d_spectral():
    return examples.gwcs_1d_spectral()


@pytest.fixture
def gwcs_spec_cel_time_4d():
    return examples.gwcs_spec_cel_time_4d()


@pytest.fixture(
    params=[
        (2, 1, 0),
        (2, 0, 1),
        pytest.param(
            (1, 0, 2), marks=pytest.mark.skip(reason="Fails round-trip for -TAB axis 3")
        ),
    ],
)
def gwcs_cube_with_separable_spectral(request):
    axes_order = request.param
    return examples.gwcs_cube_with_separable_spectral(axes_order)


@pytest.fixture(
    params=[
        (2, 0, 1),
        (2, 1, 0),
        pytest.param(
            (0, 2, 1), marks=pytest.mark.skip(reason="Fails round-trip for -TAB axis 2")
        ),
        pytest.param(
            (1, 0, 2), marks=pytest.mark.skip(reason="Fails round-trip for -TAB axis 3")
        ),
    ],
)
def gwcs_cube_with_separable_time(request):
    axes_order = request.param
    return examples.gwcs_cube_with_separable_time(axes_order)


@pytest.fixture
def gwcs_7d_complex_mapping():
    return examples.gwcs_7d_complex_mapping()


@pytest.fixture
def spher_to_cart():
    return geometry.SphericalToCartesian()


@pytest.fixture
def cart_to_spher():
    return geometry.CartesianToSpherical()


@pytest.fixture
def gwcs_with_pipeline_celestial():
    return examples.gwcs_with_pipeline_celestial()


@pytest.fixture
def gwcs_romanisim():
    return examples.gwcs_romanisim()


@pytest.fixture(
    params=[
        (5.6, -72.4),
        (5.6, 90),
        (5.6, -90),
        (5.6, 0),
    ],
)
def fits_wcs_imaging_simple(request):
    params = request.param
    return examples.fits_wcs_imaging_simple(params)


@pytest.fixture
def gwcs_2d_spatial_shift_reverse():
    return examples.gwcs_2d_spatial_shift_reverse()


@pytest.fixture
def gwcs_multi_stage():
    return examples.gwcs_multi_stage()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import io
import json
import os
import sys
import textwrap
from collections import OrderedDict

import yaml


def write_if_different(filename, data):
    """Write ``data`` to ``filename``, if the content of the file is different.

    Parameters
    ----------
    filename : str
        The file name to be written to.
    data : bytes
        The data to be written to `filename`.
    """
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))

    if os.path.exists(filename):
        with open(filename, "rb") as fd:
            original_data = fd.read()
    else:
        original_data = None

    if original_data != data:
        print(f"Converting schema {os.path.basename(filename)}")  # noqa: T201
        with open(filename, "wb") as fd:
            fd.write(data)


def write_header(o, content, level):
    """
    Write a reStructuredText header to the file.

    Parameters
    ----------
    o : output stream

    content : str
        The content of the header

    level : int
        The level of the header
    """
    levels = "=-~^."
    if level >= len(levels):
        o.write(f"**{content}**\n\n")
    else:
        o.write(content)
        o.write("\n")
        o.write(levels[level] * len(content))
        o.write("\n\n")


def format_range(
    var_middle, var_end, minimum, maximum, exclusiveMinimum, exclusiveMaximum
):
    """
    Formats an mathematical description of a range, for example, ``0 ≤
    x ≤ 2``.

    Parameters
    ----------
    var_middle : str or None
        The string to put in the middle of an expression, such as
        the ``x`` in ``0 ≤ x ≤ 2``.

    var_end : str or None
        The string to put at one end of a single comparison, such as
        the ``x`` in ``x ≤ 0``.

    minimum : number
        The minimum value.

    maximum : number
        The maximum value.

    exclusiveMinimum : bool
        If `True`, the range excludes the minimum value.

    exclusiveMaximum : bool
        If `True`, the range excludes the maximum value

    Returns
    -------
    expr : str
        The formatted range expression
    """
    if minimum is not None and maximum is not None:
        part = f"{minimum} "
        if exclusiveMinimum:
            part += "<"
        else:
            part += "≤"
        part += f" {var_middle} "
        if exclusiveMaximum:
            part += "<"
        else:
            part += "≤"
        part += f" {maximum}"
    elif minimum is not None:
        part = f"{var_end} " if var_end is not None else ""
        if exclusiveMinimum:
            part += f"> {minimum}"
        else:
            part += f"≥ {minimum}"
    elif maximum is not None:
        part = f"{var_end} " if var_end is not None else ""
        if exclusiveMaximum:
            part += f"< {maximum}"
        else:
            part += f"≤ {maximum}"
    else:
        return None
    return part


def format_type(schema, root):
    """
    Creates an English/mathematical description of a schema fragment.

    Parameters
    ----------
    schema : JSON schema fragment

    root : str
        The JSON path to the schema fragment.
    """
    if "anyOf" in schema:
        return " :soft:`or` ".join(format_type(x, root) for x in schema["anyOf"])

    if "allOf" in schema:
        return " :soft:`and` ".join(format_type(x, root) for x in schema["allOf"])

    if "$ref" in schema:
        ref = schema["$ref"]
        if ref.startswith("#/"):
            return f":ref:`{ref[2:]} <{root}/{ref[2:]}>`"
        basename = os.path.basename(ref)
        if "tag:stsci.edu:asdf" in ref or "tag:astropy.org:astropy" in ref:
            return f"`{basename} <{ref}>`"
        return f":doc:`{basename} <{ref}>`"

    type = schema.get("type")  # noqa: A001
    if isinstance(type, list):
        parts = [" or ".join(type)]

    elif type is None:
        parts = ["any"]

    else:
        parts = [type]

    if type == "string":
        range = format_range(  # noqa: A001
            "*len*",
            "*len*",
            schema.get("minLength"),
            schema.get("maxLength"),
            False,
            False,
        )
        if range is not None or "pattern" in schema or "format" in schema:
            parts.append("(")
            if range is not None:
                parts.append(range)
            if "pattern" in schema:
                pattern = schema["pattern"].encode("unicode_escape")
                pattern = pattern.decode("ascii")
                parts.append(f":soft:`regex` :regexp:`{pattern}`")
            if "format" in schema:
                parts.append(":soft:`format` {}".format(schema["format"]))
            parts.append(")")

    elif type in ("integer", "number"):
        range = format_range(  # noqa: A001
            "*x*",
            "",
            schema.get("minimum"),
            schema.get("maximum"),
            schema.get("exclusiveMinimum"),
            schema.get("exclusiveMaximum"),
        )
        if range is not None:
            parts.append(range)
        # TODO: multipleOf

    elif type == "object":
        range = format_range(  # noqa: A001
            "*len*",
            "*len*",
            schema.get("minProperties"),
            schema.get("maxProperties"),
            False,
            False,
        )
        if range is not None:
            parts.append(range)
        # TODO: Dependencies
        # TODO: Pattern properties

    elif type == "array":
        items = schema.get("items")
        if schema.get("items") and isinstance(items, dict):
            if schema.get("uniqueItems"):
                parts.append(":soft:`of unique`")
            else:
                parts.append(":soft:`of`")
            parts.append("(")
            parts.append(format_type(items, root))
            parts.append(")")
        range = format_range(  # noqa: A001
            "*len*",
            "*len*",
            schema.get("minItems"),
            schema.get("maxItems"),
            False,
            False,
        )
        if range is not None:
            parts.append(range)

    if "enum" in schema:
        parts.append(":soft:`from`")
        parts.append(json.dumps(schema["enum"]))

    return " ".join(parts)


def reindent(content, indent):
    """
    Reindent a string to the given number of spaces.
    """
    content = textwrap.dedent(content)
    lines = [indent + line for line in content.split("\n")]
    return "\n".join(lines)


def recurse(o, name, schema, path, level, required=False):
    """
    Convert a schema fragment to reStructuredText.

    Parameters
    ----------
    o : output stream

    name : str
        Name of the entry

    schema : schema fragment

    path : list of str
        Path to schema fragment

    level : int
        Indentation level

    required : bool
        If `True` the entry is required by the schema and will be
        documented as such.
    """
    indent = "  " * max(level, 0)
    o.write("\n\n")
    o.write(indent)
    o.write(f".. _{os.path.join(*path)}:\n\n")
    if level == 0:
        write_header(o, name, level)
    elif name != "items":
        o.write(indent)
        o.write(f":entry:`{name}`\n\n")

    o.write(indent)
    if path[0].startswith("tag:stsci.edu:asdf"):
        o.write(format_type(schema, path[0]))
    else:
        o.write(":soft:`Type:` ")
        o.write(format_type(schema, path[0]))
    o.write(".")
    if required:
        o.write(" Required.")
    o.write("\n\n")

    o.write(reindent(schema.get("title", ""), indent))
    o.write("\n\n")

    o.write(reindent(schema.get("description", ""), indent))
    o.write("\n\n")

    if "default" in schema:
        o.write(indent)
        o.write(":soft:`Default:` {}".format(json.dumps(schema["default"])))
        o.write("\n\n")

    if "definitions" in schema:
        o.write(indent)
        o.write(":category:`Definitions:`\n\n")
        for key, val in schema["definitions"].items():
            recurse(o, key, val, [*path, "definitions", key], level + 1)

    if "anyOf" in schema and len(schema["anyOf"]) > 1:
        o.write(indent)
        o.write(":category:`Any of:`\n\n")
        for i, subschema in enumerate(schema["anyOf"]):
            recurse(o, "—", subschema, [*path, "anyOf", str(i)], level + 1)

    elif "allOf" in schema and len(schema["allOf"]) > 1:
        o.write(indent)
        o.write(":category:`All of:`\n\n")
        for i, subschema in enumerate(schema["allOf"]):
            recurse(o, i, subschema, [*path, "allOf", str(i)], level + 1)

    if schema.get("type") == "object":
        o.write(indent)
        o.write(":category:`Properties:`\n\n")
        for key, val in schema.get("properties", {}).items():
            recurse(
                o,
                key,
                val,
                [*path, "properties", key],
                level + 1,
                key in schema.get("required", []),
            )

    elif schema.get("type") == "array":
        o.write(indent)
        o.write(":category:`Items:`\n\n")
        items = schema.get("items")
        if isinstance(items, dict):
            recurse(o, "items", items, [*path, "items"], level + 1)
        elif isinstance(items, list):
            for i, val in enumerate(items):
                name = f"index[{i}]"
                recurse(o, name, val, [*path, str(i)], level + 1)

    if "examples" in schema:
        o.write(indent)
        o.write(":category:`Examples:`\n\n")
        for description, example in schema["examples"]:
            o.write(reindent(description + "::\n\n", indent))
            o.write(reindent(example, indent + "  "))
            o.write("\n\n")


def convert_schema_to_rst(src, dst):
    """
    Convert a YAML schema to reStructuredText.
    """
    with open(src, "rb") as fd:
        schema = yaml.safe_load(fd)
    with open(src, "rb") as fd:
        yaml_content = fd.read()

    o = io.StringIO()

    id = schema.get("id", "#")  # noqa: A001
    name = os.path.basename(src[:-5])
    if "title" in schema:
        name += ": " + schema["title"].strip()
    recurse(o, name, schema, [id], 0)

    # o.write(".. only:: html\n\n   :download:`Original schema in YAML <{0}>`\n".
    # os.path.basename(src)))

    write_if_different(dst, yaml_content)
    write_if_different(dst[:-5] + ".rst", o.getvalue().encode("utf-8"))


def construct_mapping(self, node, deep=False):
    """
    Make sure the properties are written out in the same order as the
    original file.
    """
    if not isinstance(node, yaml.MappingNode):
        raise yaml.constructor.ConstructorError(
            None,
            None,
            f"expected a mapping node, but found {node.id}",
            node.start_mark,
        )
    mapping = OrderedDict()
    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        try:
            hash(key)
        except TypeError as exc:
            msg = "while constructing a mapping"
            raise yaml.constructor.ConstructorError(
                msg,
                node.start_mark,
                f"found unacceptable key ({exc})",
                key_node.start_mark,
            ) from exc
        value = self.construct_object(value_node, deep=deep)
        mapping[key] = value
    return mapping


yaml.SafeLoader.add_constructor("tag:yaml.org,2002:map", construct_mapping)


def main(src, dst):
    for root, _, files in os.walk(src):
        for fname in files:
            if not fname.endswith(".yaml"):
                continue
            src_path = os.path.join(root, fname)
            dst_path = os.path.join(dst, os.path.relpath(src_path, src))

            convert_schema_to_rst(src_path, dst_path)


def decode_filename(fname):
    return fname


if __name__ == "__main__":
    src = decode_filename(sys.argv[-2])
    dst = decode_filename(sys.argv[-1])

    sys.exit(main(src, dst))



from ._core import CoordinateFrame

__all__ = ["EmptyFrame"]


class EmptyFrame(CoordinateFrame):
    """
    Represents a "default" detector frame. This is for use as the default value
    for input frame by the WCS object.
    """

    def __init__(self, name=None):
        self._name = "detector" if name is None else name

    def __repr__(self):
        return f'<{type(self).__name__}(name="{self.name}")>'

    def __str__(self):
        if self._name is not None:
            return self._name
        return type(self).__name__

    @property
    def name(self):
        """A custom name of this frame."""
        return self._name

    @name.setter
    def name(self, val):
        """A custom name of this frame."""
        self._name = val

    def _raise_error(self) -> None:
        msg = "EmptyFrame does not have any information"
        raise NotImplementedError(msg)

    @property
    def naxes(self):
        self._raise_error()

    @property
    def unit(self):
        return None

    @property
    def axes_names(self):
        self._raise_error()

    @property
    def axes_order(self):
        self._raise_error()

    @property
    def reference_frame(self):
        self._raise_error()

    @property
    def axes_type(self):
        self._raise_error()

    @property
    def axis_physical_types(self):
        self._raise_error()

    @property
    def world_axis_object_classes(self):
        self._raise_error()

    @property
    def _native_world_axis_object_components(self):
        self._raise_error()

    def to_high_level_coordinates(self, *values):
        self._raise_error()

    def from_high_level_coordinates(self, *high_level_coords):
        self._raise_error()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Spectroscopy related models.
"""

import astropy.units as u
import numpy as np
from astropy.modeling.core import Model
from astropy.modeling.parameters import Parameter

__all__ = [
    "AnglesFromGratingEquation3D",
    "SellmeierGlass",
    "SellmeierZemax",
    "Snell3D",
    "WavelengthFromGratingEquation",
]


class WavelengthFromGratingEquation(Model):
    r"""Solve the Grating Dispersion Law for the wavelength.

    .. Note:: This form of the equation can be used for paraxial
      (small angle approximation) as well as oblique incident angles.
      With paraxial systems the inputs are ``sin`` of the angles and it
      transforms to

      :math:`(\sin(\alpha_{in}) + \sin(\alpha_{out})) / (groove\_density * spectral\_order)`.

      With oblique angles the inputs are the direction cosines
      of the angles.


    Parameters
    ----------
    groove_density : int
        Grating ruling density in units of 1/length.
    spectral_order : int
        Spectral order.

    Examples
    --------
    >>> from astropy.modeling.models import math
    >>> model = WavelengthFromGratingEquation(groove_density=20000*1/u.m, spectral_order=-1)
    >>> alpha_in = (math.Deg2radUfunc() | math.SinUfunc())(.0001 * u.deg)
    >>> alpha_out = (math.Deg2radUfunc() | math.SinUfunc())(.0001 * u.deg)
    >>> lam = model(alpha_in, alpha_out)
    >>> print(lam)
    -1.7453292519934437e-10 m

    """  # noqa: E501

    _separable = False

    linear = False

    n_inputs = 2
    n_outputs = 1

    groove_density = Parameter(default=1)
    """ Grating ruling density in units of 1/m."""
    spectral_order = Parameter(default=1)
    """ Spectral order."""

    def __init__(self, groove_density, spectral_order, **kwargs):
        super().__init__(
            groove_density=groove_density, spectral_order=spectral_order, **kwargs
        )
        self.inputs = ("alpha_in", "alpha_out")
        """ Sine function of the angles or the direction cosines."""
        self.outputs = ("wavelength",)
        """ Wavelength."""

    def evaluate(self, alpha_in, alpha_out, groove_density, spectral_order):
        return (alpha_in + alpha_out) / (groove_density * spectral_order)

    @property
    def return_units(self):
        if self.groove_density.unit is None:
            return None
        return {"wavelength": u.Unit(1 / self.groove_density.unit)}


class AnglesFromGratingEquation3D(Model):
    """
    Solve the 3D Grating Dispersion Law in Direction Cosine
    space for the refracted angle.

    Parameters
    ----------
    groove_density : int
        Grating ruling density in units of 1/m.
    order : int
        Spectral order.

    Examples
    --------
    >>> from astropy.modeling.models import math
    >>> model = AnglesFromGratingEquation3D(groove_density=20000*1/u.m, spectral_order=-1)
    >>> alpha_in = (math.Deg2radUfunc() | math.SinUfunc())(.0001 * u.deg)
    >>> beta_in = (math.Deg2radUfunc() | math.SinUfunc())(.0001 * u.deg)
    >>> lam = 2e-6 * u.m
    >>> alpha_out, beta_out, gamma_out = model(lam, alpha_in, beta_in)
    >>> print(alpha_out, beta_out, gamma_out)
    0.04000174532925199 -1.7453292519934436e-06 0.9991996098716049

    """  # noqa: E501

    _separable = False
    linear = False

    n_inputs = 3
    n_outputs = 3

    groove_density = Parameter(default=1)
    """ Grating ruling density in units 1/ length."""

    spectral_order = Parameter(default=1)
    """ Spectral order."""

    def __init__(self, groove_density, spectral_order, **kwargs):
        super().__init__(
            groove_density=groove_density, spectral_order=spectral_order, **kwargs
        )
        self.inputs = ("wavelength", "alpha_in", "beta_in")
        """ Wavelength and 2 angle coordinates going into the grating."""

        self.outputs = ("alpha_out", "beta_out", "gamma_out")
        """ Two angles coming out of the grating. """

    def evaluate(self, wavelength, alpha_in, beta_in, groove_density, spectral_order):
        if alpha_in.shape != beta_in.shape:
            msg = "Expected input arrays to have the same shape."
            raise ValueError(msg)

        if isinstance(groove_density, u.Quantity):
            alpha_in = u.Quantity(alpha_in)
            beta_in = u.Quantity(beta_in)

        alpha_out = -groove_density * spectral_order * wavelength + alpha_in
        beta_out = -beta_in
        gamma_out = np.sqrt(1 - alpha_out**2 - beta_out**2)
        return alpha_out, beta_out, gamma_out

    @property
    def input_units(self):
        if self.groove_density.unit is None:
            return None
        return {
            "wavelength": 1 / self.groove_density.unit,
            "alpha_in": u.Unit(1),
            "beta_in": u.Unit(1),
        }


class Snell3D(Model):
    """
    Snell model in 3D form.

    Inputs are index of refraction and direction cosines.

    Returns
    -------
    alpha_out, beta_out, gamma_out : float
        Direction cosines.
    """

    _separable = False
    linear = False

    n_inputs = 4
    n_outputs = 3

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.inputs = ("n", "alpha_in", "beta_in", "gamma_in")
        self.outputs = ("alpha_out", "beta_out", "gamma_out")

    @staticmethod
    def evaluate(n, alpha_in, beta_in, gamma_in):
        # Apply Snell's law through front surface,
        # eq 5.3.3 II in Nirspec docs
        alpha_out = alpha_in / n
        beta_out = beta_in / n
        gamma_out = np.sqrt(1.0 - alpha_out**2 - beta_out**2)
        return alpha_out, beta_out, gamma_out


class SellmeierGlass(Model):
    """
    Sellmeier equation for glass.

    Parameters
    ----------
    B_coef : ndarray
        Iterable of size 3 containing B coefficients.
    C_coef : ndarray
        Iterable of size 3 containing c coefficients in
        units of ``u.um**2``.

    Returns
    -------
    n : float
        Refractive index.

    Examples
    --------
    >>> import astropy.units as u
    >>> b_coef = [0.58339748, 0.46085267, 3.8915394]
    >>> c_coef = [0.00252643, 0.010078333, 1200.556] * u.um**2
    >>> model = SellmeierGlass(b_coef, c_coef)
    >>> model(2 * u.m)
        <Quantity 2.43634758>

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Sellmeier_equation

    Notes
    -----

    Model formula:

        .. math::

            n(\\lambda)^2 = 1 + \\frac{(B1 * \\lambda^2 )}{(\\lambda^2 - C1)} +
            \\frac{(B2 * \\lambda^2 )}{(\\lambda^2 - C2)} +
            \\frac{(B3 * \\lambda^2 )}{(\\lambda^2 - C3)}

    """

    _separable = False
    standard_broadcasting = False
    linear = False

    n_inputs = 1
    n_outputs = 1

    B_coef = Parameter(default=np.array([1, 1, 1]))
    """ B1, B2, B3 coefficients. """
    C_coef = Parameter(default=np.array([0, 0, 0]))
    """ C1, C2, C3 coefficients in units of um ** 2. """

    def __init__(self, B_coef, C_coef, **kwargs):
        super().__init__(B_coef, C_coef)
        self.inputs = ("wavelength",)
        self.outputs = ("n",)

    @staticmethod
    def evaluate(wavelength, B_coef, C_coef):
        B1, B2, B3 = B_coef[0]
        C1, C2, C3 = C_coef[0]

        return np.sqrt(
            1.0
            + B1 * wavelength**2 / (wavelength**2 - C1)
            + B2 * wavelength**2 / (wavelength**2 - C2)
            + B3 * wavelength**2 / (wavelength**2 - C3)
        )

    @property
    def input_units(self):
        if self.C_coef.unit is None:
            return None
        return {"wavelength": u.um}


class SellmeierZemax(Model):
    """
    Sellmeier equation used by Zemax.

    Parameters
    ----------
    temperature : float
        Temperature of the material in ``u.Kelvin``.
    ref_temperature : float
        Reference emperature of the glass in ``u.Kelvin``.
    ref_pressure : float
        Reference pressure in ATM.
    pressure : float
        Measured pressure in ATM.
    B_coef : ndarray
        Iterable of size 3 containing B coefficients.
    C_coef : ndarray
        Iterable of size 3 containing C coefficients in
        units of ``u.um**2``.
    D_coef : ndarray
        Iterable of size 3 containing constants to describe the
        behavior of the material.
    E_coef : ndarray
        Iterable of size 3 containing constants to describe the
        behavior of the material.

    Returns
    -------
    n : float
        Refractive index.

    """

    _separable = False
    standard_broadcasting = False
    linear = False

    n_inputs = 1
    n_outputs = 1

    temperature = Parameter(default=0)
    ref_temperature = Parameter(default=0)
    ref_pressure = Parameter(default=0)
    pressure = Parameter(default=0)
    B_coef = Parameter(default=[1, 1, 1])
    C_coef = Parameter(default=[0, 0, 0])
    D_coef = Parameter(default=[0, 0, 0])
    E_coef = Parameter(default=[1, 1, 1])

    def __init__(
        self,
        temperature=temperature,
        ref_temperature=ref_temperature,
        ref_pressure=ref_pressure,
        pressure=pressure,
        B_coef=B_coef,
        C_coef=C_coef,
        D_coef=D_coef,
        E_coef=E_coef,
        **kwargs,
    ):
        super().__init__(
            temperature=temperature,
            ref_temperature=ref_temperature,
            ref_pressure=ref_pressure,
            pressure=pressure,
            B_coef=B_coef,
            C_coef=C_coef,
            D_coef=D_coef,
            E_coef=E_coef,
            **kwargs,
        )
        self.inputs = ("wavelength",)
        self.outputs = ("n",)

    def evaluate(
        self,
        wavelength,
        temp,
        ref_temp,
        ref_pressure,
        pressure,
        B_coef,
        C_coef,
        D_coef,
        E_coef,
    ):
        """
        Input ``wavelength`` is in units of microns.
        """
        if isinstance(temp, u.Quantity):
            temp = temp.to(u.Celsius)
            ref_temp = ref_temp.to(u.Celsius)
        else:
            KtoC = 273.15  # kelvin to celsius conversion
            temp -= KtoC
            ref_temp -= KtoC
        delta = temp - ref_temp
        D0, D1, D2 = D_coef[0]
        E0, E1, lam_tk = E_coef[0]

        nref = (
            1.0
            + (
                6432.8
                + 2949810.0 * wavelength**2 / (146.0 * wavelength**2 - 1.0)
                + (5540.0 * wavelength**2) / (41.0 * wavelength**2 - 1.0)
            )
            * 1e-8
        )
        # T should be in C, P should be in ATM
        nair_obs = 1.0 + ((nref - 1.0) * pressure) / (1.0 + (temp - 15.0) * 3.4785e-3)
        nair_ref = 1.0 + ((nref - 1.0) * ref_pressure) / (
            1.0 + (ref_temp - 15) * 3.4785e-3
        )

        # Compute the relative index of the glass at Tref and Pref using
        # Sellmeier equation I.
        lamrel = wavelength * nair_obs / nair_ref
        nrel = SellmeierGlass.evaluate(lamrel, B_coef, C_coef)
        # Convert the relative index of refraction at the reference temperature
        # and pressure to absolute.
        nabs_ref = nrel * nair_ref

        # Compute the absolute index of the glass
        delnabs = (0.5 * (nrel**2 - 1.0) / nrel) * (
            D0 * delta
            + D1 * delta**2
            + D2 * delta**3
            + (E0 * delta + E1 * delta**2) / (lamrel**2 - lam_tk**2)
        )
        nabs_obs = nabs_ref + delnabs

        # Define the relative index at the system's operating T and P.
        return nabs_obs / nair_obs



from collections import defaultdict

import numpy as np

from ._core import CoordinateFrame

__all__ = ["CompositeFrame"]


class CompositeFrame(CoordinateFrame):
    """
    Represents one or more frames.

    Parameters
    ----------
    frames : list
        List of constituient frames.
    name : str
        Name for this frame.
    """

    def __init__(self, frames, name=None):
        self._frames = frames[:]
        naxes = sum([frame._naxes for frame in self._frames])

        axes_order = []
        axes_type = []
        axes_names = []
        unit = []
        ph_type = []

        for frame in frames:
            axes_order.extend(frame.axes_order)

        # Stack the raw (not-native) ordered properties
        for frame in frames:
            axes_type += list(frame._prop.axes_type)
            axes_names += list(frame._prop.axes_names)
            unit += list(frame._prop.unit)
            ph_type += list(frame._prop.axis_physical_types)

        if len(np.unique(axes_order)) != len(axes_order):
            msg = (
                "Incorrect numbering of axes, "
                "axes_order should contain unique numbers, "
                f"got {axes_order}."
            )
            raise ValueError(msg)

        super().__init__(
            naxes,
            axes_type=axes_type,
            axes_order=axes_order,
            unit=unit,
            axes_names=axes_names,
            axis_physical_types=tuple(ph_type),
            name=name,
        )
        self._axis_physical_types = tuple(ph_type)

    @property
    def frames(self):
        """
        The constituient frames that comprise this `CompositeFrame`.
        """
        return self._frames

    def __repr__(self):
        return repr(self.frames)

    @property
    def _wao_classes_rename_map(self):
        mapper = defaultdict(dict)
        seen_names = []
        for frame in self.frames:
            # ensure the frame is in the mapper
            mapper[frame]
            for key in frame.world_axis_object_classes:
                if key in seen_names:
                    new_key = f"{key}{seen_names.count(key)}"
                    mapper[frame][key] = new_key
                seen_names.append(key)
        return mapper

    @property
    def _wao_renamed_components_iter(self):
        mapper = self._wao_classes_rename_map
        for frame in self.frames:
            renamed_components = []
            for component in frame._native_world_axis_object_components:
                comp = list(component)
                rename = mapper[frame].get(comp[0])
                if rename:
                    comp[0] = rename
                renamed_components.append(tuple(comp))
            yield frame, renamed_components

    @property
    def _wao_renamed_classes_iter(self):
        mapper = self._wao_classes_rename_map
        for frame in self.frames:
            for key, value in frame.world_axis_object_classes.items():
                rename = mapper[frame].get(key)
                yield rename if rename else key, value

    @property
    def world_axis_object_components(self):
        out = [None] * self.naxes

        for frame, components in self._wao_renamed_components_iter:
            for i, ao in enumerate(frame.axes_order):
                out[ao] = components[i]

        if any(o is None for o in out):
            msg = "axes_order leads to incomplete world_axis_object_components"
            raise ValueError(msg)

        return out

    @property
    def world_axis_object_classes(self):
        return dict(self._wao_renamed_classes_iter)



import astropy.units as u
from astropy.coordinates import Galactic, SkyCoord, SpectralCoord
from astropy.wcs.wcsapi import wcs_info_str
from astropy.wcs.wcsapi.wrappers import SlicedLowLevelWCS
from numpy.testing import assert_allclose, assert_equal

EXPECTED_ELLIPSIS_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (30, 20, 10)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              10  (-1, 35)
        1  None              20  (-2, 45)
        2  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  em.freq           Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no  yes
        1   no  yes   no
        2  yes   no  yes
"""


def test_no_ellipsis(gwcs_3d_galactic_spectral):
    expected_repr = EXPECTED_ELLIPSIS_REPR.replace("SlicedLowLevel", "")
    assert wcs_info_str(gwcs_3d_galactic_spectral) == expected_repr.strip()


def test_ellipsis(gwcs_3d_galactic_spectral):
    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, Ellipsis)

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert wcs.array_shape == (30, 20, 10)
    assert wcs.pixel_shape == (10, 20, 30)
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        "em.freq",
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix,
        [[True, False, True], [False, True, False], [True, False, True]],
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert wcs.world_axis_object_classes["spectral"][0] is SpectralCoord
    assert wcs.world_axis_object_classes["spectral"][1] == ()
    assert wcs.world_axis_object_classes["spectral"][2] == {"unit": "Hz"}

    assert_allclose(wcs.pixel_to_world_values(29, 39, 44), (80, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 39, 29), (80, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 20, 205), (29.0, 39.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 20, 205), (44, 39, 29))

    assert str(wcs) == EXPECTED_ELLIPSIS_REPR.strip()
    assert EXPECTED_ELLIPSIS_REPR.strip() in repr(wcs)

    assert_equal(wcs.pixel_bounds, [(-1, 35), (-2, 45), (5, 50)])


EXPECTED_SPECTRAL_SLICE_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 2 pixel and 2 world dimensions

Array shape (Numpy order): (30, 10)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              10  (-1, 35)
        1  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0  yes  yes
        1  yes  yes
"""


def test_spectral_slice(gwcs_3d_galactic_spectral):
    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, [slice(None), 10])

    assert wcs.pixel_n_dim == 2
    assert wcs.world_n_dim == 2
    assert wcs.array_shape == (30, 10)
    assert wcs.pixel_shape == (10, 30)
    assert wcs.world_axis_physical_types == ["pos.galactic.lat", "pos.galactic.lon"]
    assert wcs.world_axis_units == ["deg", "deg"]

    assert_equal(wcs.axis_correlation_matrix, [[True, True], [True, True]])

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert_allclose(wcs.pixel_to_world_values(29, 44), (80, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 29), (80, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 205), (29.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 205), (44, 29))

    assert_equal(wcs.pixel_bounds, [(-1, 35), (5, 50)])

    assert str(wcs) == EXPECTED_SPECTRAL_SLICE_REPR.strip()
    assert EXPECTED_SPECTRAL_SLICE_REPR.strip() in repr(wcs)


EXPECTED_SPECTRAL_RANGE_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (30, 6, 10)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              10  (-1, 35)
        1  None               6  (-6, 41)
        2  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  em.freq           Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no  yes
        1   no  yes   no
        2  yes   no  yes
"""


def test_spectral_range(gwcs_3d_galactic_spectral):
    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, [slice(None), slice(4, 10)])

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert wcs.array_shape == (30, 6, 10)
    assert wcs.pixel_shape == (10, 6, 30)
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        "em.freq",
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix,
        [[True, False, True], [False, True, False], [True, False, True]],
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert wcs.world_axis_object_classes["spectral"][0] is SpectralCoord
    assert wcs.world_axis_object_classes["spectral"][1] == ()
    assert wcs.world_axis_object_classes["spectral"][2] == {"unit": "Hz"}

    assert_allclose(wcs.pixel_to_world_values(29, 35, 44), (80, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 35, 29), (80, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 20, 205), (29.0, 35.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 20, 205), (44, 35, 29))

    assert_equal(wcs.pixel_bounds, [(-1, 35), (-6, 41), (5, 50)])

    assert str(wcs) == EXPECTED_SPECTRAL_RANGE_REPR.strip()
    assert EXPECTED_SPECTRAL_RANGE_REPR.strip() in repr(wcs)


EXPECTED_CELESTIAL_SLICE_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 2 pixel and 3 world dimensions

Array shape (Numpy order): (30, 20)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              20  (-2, 45)
        1  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  em.freq           Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0   no  yes
        1  yes   no
        2   no  yes
"""


def test_celestial_slice(gwcs_3d_galactic_spectral):
    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, [Ellipsis, 5])

    assert wcs.pixel_n_dim == 2
    assert wcs.world_n_dim == 3
    assert wcs.array_shape == (30, 20)
    assert wcs.pixel_shape == (20, 30)
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        "em.freq",
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix, [[False, True], [True, False], [False, True]]
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert wcs.world_axis_object_classes["spectral"][0] is SpectralCoord
    assert wcs.world_axis_object_classes["spectral"][1] == ()
    assert wcs.world_axis_object_classes["spectral"][2] == {"unit": "Hz"}

    assert_allclose(wcs.pixel_to_world_values(39, 44), (79.76, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 39), (79.76, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(79.76, 20, 205), (39.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(79.76, 20, 205), (44, 39))

    assert_equal(wcs.pixel_bounds, [(-2, 45), (5, 50)])

    assert str(wcs) == EXPECTED_CELESTIAL_SLICE_REPR.strip()
    assert EXPECTED_CELESTIAL_SLICE_REPR.strip() in repr(wcs)


EXPECTED_CELESTIAL_RANGE_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (30, 20, 5)

Pixel Dim  Axis Name  Data size  Bounds
        0  None               5  (-6, 30)
        1  None              20  (-2, 45)
        2  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  em.freq           Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no  yes
        1   no  yes   no
        2  yes   no  yes
"""


def test_celestial_range(gwcs_3d_galactic_spectral):
    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, [Ellipsis, slice(5, 10)])

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert wcs.array_shape == (30, 20, 5)
    assert wcs.pixel_shape == (5, 20, 30)
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        "em.freq",
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix,
        [[True, False, True], [False, True, False], [True, False, True]],
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert wcs.world_axis_object_classes["spectral"][0] is SpectralCoord
    assert wcs.world_axis_object_classes["spectral"][1] == ()
    assert wcs.world_axis_object_classes["spectral"][2] == {"unit": "Hz"}

    assert_allclose(wcs.pixel_to_world_values(24, 39, 44), (80, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 39, 24), (80, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 20, 205), (24.0, 39.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 20, 205), (44, 39, 24))

    assert_equal(wcs.pixel_bounds, [(-6, 30), (-2, 45), (5, 50)])

    assert str(wcs) == EXPECTED_CELESTIAL_RANGE_REPR.strip()
    assert EXPECTED_CELESTIAL_RANGE_REPR.strip() in repr(wcs)


EXPECTED_NO_SHAPE_REPR = """
SlicedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): None

Pixel Dim  Axis Name  Data size  Bounds
        0  None            None  None
        1  None            None  None
        2  None            None  None

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  em.freq           Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no  yes
        1   no  yes   no
        2  yes   no  yes
"""


def test_no_array_shape(gwcs_3d_galactic_spectral):
    gwcs_3d_galactic_spectral.pixel_shape = None
    gwcs_3d_galactic_spectral.array_shape = None
    gwcs_3d_galactic_spectral.forward_transform.bounding_box = None

    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, Ellipsis)

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert wcs.array_shape is None
    assert wcs.pixel_shape is None
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        "em.freq",
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix,
        [[True, False, True], [False, True, False], [True, False, True]],
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert wcs.world_axis_object_classes["spectral"][0] is SpectralCoord
    assert wcs.world_axis_object_classes["spectral"][1] == ()
    assert wcs.world_axis_object_classes["spectral"][2] == {"unit": "Hz"}

    assert_allclose(wcs.pixel_to_world_values(29, 39, 44), (80, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 39, 29), (80, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 20, 205), (29.0, 39.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 20, 205), (44, 39, 29))

    assert str(wcs) == EXPECTED_NO_SHAPE_REPR.strip()
    assert EXPECTED_NO_SHAPE_REPR.strip() in repr(wcs)


# Testing the WCS object having some physical types as None/Unknown
EXPECTED_ELLIPSIS_REPR_NONE_TYPES = """
SlicedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (30, 20, 10)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              10  (-1, 35)
        1  None              20  (-2, 45)
        2  None              30  (5, 50)

World Dim  Axis Name  Physical Type     Units
        0  Latitude   pos.galactic.lat  deg
        1  Frequency  None              Hz
        2  Longitude  pos.galactic.lon  deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0  yes   no  yes
        1   no  yes   no
        2  yes   no  yes
"""


def test_ellipsis_none_types(gwcs_3d_galactic_spectral):
    pht = list(gwcs_3d_galactic_spectral.output_frame._axis_physical_types)
    # This index is in "axes_order" ordering
    pht[2] = None
    gwcs_3d_galactic_spectral.output_frame._prop.axis_physical_types = tuple(pht)

    wcs = SlicedLowLevelWCS(gwcs_3d_galactic_spectral, Ellipsis)

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert wcs.array_shape == (30, 20, 10)
    assert wcs.pixel_shape == (10, 20, 30)
    assert wcs.world_axis_physical_types == [
        "pos.galactic.lat",
        None,
        "pos.galactic.lon",
    ]
    assert wcs.world_axis_units == ["deg", "Hz", "deg"]

    assert_equal(
        wcs.axis_correlation_matrix,
        [[True, False, True], [False, True, False], [True, False, True]],
    )

    first_two = [c[:2] for c in wcs.world_axis_object_components]
    last_one = [c[2] for c in wcs.world_axis_object_components]
    assert first_two == [("celestial", 1), ("spectral", 0), ("celestial", 0)]

    assert all(callable(last) for last in last_one)

    assert wcs.world_axis_object_classes["celestial"][0] is SkyCoord
    assert wcs.world_axis_object_classes["celestial"][1] == ()
    assert isinstance(wcs.world_axis_object_classes["celestial"][2]["frame"], Galactic)
    assert tuple(wcs.world_axis_object_classes["celestial"][2]["unit"]) == (
        u.deg,
        u.deg,
    )

    assert_allclose(wcs.pixel_to_world_values(29, 39, 44), (80, 20, 205))
    assert_allclose(wcs.array_index_to_world_values(44, 39, 29), (80, 20, 205))

    assert_allclose(wcs.world_to_pixel_values(80, 20, 205), (29.0, 39.0, 44.0))
    assert_equal(wcs.world_to_array_index_values(80, 20, 205), (44, 39, 29))

    assert_equal(wcs.pixel_bounds, [(-1, 35), (-2, 45), (5, 50)])

    assert str(wcs) == EXPECTED_ELLIPSIS_REPR_NONE_TYPES.strip()
    assert EXPECTED_ELLIPSIS_REPR_NONE_TYPES.strip() in repr(wcs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from astropy import units as u

from ._core import CoordinateFrame

__all__ = ["Frame2D"]


class Frame2D(CoordinateFrame):
    """
    A 2D coordinate frame.

    Parameters
    ----------
    axes_order : tuple of int
        A dimension in the input data that corresponds to this axis.
    unit : list of astropy.units.Unit
        Unit for each axis.
    axes_names : list
        Names of the axes in this frame.
    name : str
        Name of this frame.
    """

    def __init__(
        self,
        axes_order=(0, 1),
        unit=(u.pix, u.pix),
        axes_names=("x", "y"),
        name=None,
        axes_type=None,
        axis_physical_types=None,
    ):
        if axes_type is None:
            axes_type = ["SPATIAL", "SPATIAL"]
        pht = axis_physical_types or self._default_axis_physical_types(
            axes_names, axes_type
        )

        super().__init__(
            naxes=2,
            axes_type=axes_type,
            axes_order=axes_order,
            name=name,
            axes_names=axes_names,
            unit=unit,
            axis_physical_types=pht,
        )

    def _default_axis_physical_types(self, axes_names, axes_type):
        if axes_names is not None and all(axes_names):
            ph_type = axes_names
        else:
            ph_type = axes_type

        return tuple(f"custom:{t}" for t in ph_type)



import warnings

from astropy.wcs.wcsapi.fitswcs import CTYPE_TO_UCD1

__all__ = ["get_ctype_from_ucd"]


def _ucd1_to_ctype_name_mapping(ctype_to_ucd, allowed_ucd_duplicates):
    inv_map = {}
    new_ucd = set()

    for kwd, ucd in ctype_to_ucd.items():
        if ucd in inv_map:
            if ucd not in allowed_ucd_duplicates:
                new_ucd.add(ucd)
            continue
        inv_map[ucd] = allowed_ucd_duplicates.get(ucd, kwd)

    if new_ucd:
        warnings.warn(
            "Found unsupported duplicate physical type in 'astropy' mapping to CTYPE.\n"
            "Update 'gwcs' to the latest version or notify 'gwcs' developer.\n"
            "Duplicate physical types will be mapped to the following CTYPEs:\n"
            + "\n".join([f"{ucd!r:s} --> {inv_map[ucd]!r:s}" for ucd in new_ucd])
        )

    return inv_map


# List below allowed physical type duplicates and a corresponding CTYPE
# to which all duplicates will be mapped to:
_ALLOWED_UCD_DUPLICATES = {
    "time": "TIME",
    "em.wl": "WAVE",
}

UCD1_TO_CTYPE = _ucd1_to_ctype_name_mapping(
    ctype_to_ucd=CTYPE_TO_UCD1, allowed_ucd_duplicates=_ALLOWED_UCD_DUPLICATES
)


def get_ctype_from_ucd(ucd):
    """
    Return the FITS ``CTYPE`` corresponding to a UCD1 value.

    Parameters
    ----------
    ucd : str
        UCD string, for example one of ```WCS.world_axis_physical_types``.

    Returns
    -------
    CTYPE : str
        The corresponding FITS ``CTYPE`` value or an empty string.
    """
    return UCD1_TO_CTYPE.get(ucd, "")



from astropy import units as u
from astropy.coordinates import StokesCoord

from ._core import CoordinateFrame

__all__ = ["StokesFrame"]


class StokesFrame(CoordinateFrame):
    """
    A coordinate frame for representing Stokes polarisation states.

    Parameters
    ----------
    name : str
        Name of this frame.
    axes_order : tuple
        A dimension in the data that corresponds to this axis.
    """

    def __init__(
        self,
        axes_order=(0,),
        axes_names=("stokes",),
        name=None,
        axis_physical_types=None,
    ):
        pht = axis_physical_types or self._default_axis_physical_types()

        super().__init__(
            1,
            ["STOKES"],
            axes_order,
            name=name,
            axes_names=axes_names,
            unit=u.one,
            axis_physical_types=pht,
        )

    def _default_axis_physical_types(self):
        return ("phys.polarization.stokes",)

    @property
    def world_axis_object_classes(self):
        return {
            "stokes": (
                StokesCoord,
                (),
                {},
            )
        }

    @property
    def _native_world_axis_object_components(self):
        return [("stokes", 0, "value")]



import warnings

import numpy as np
from astropy import coordinates as coord
from astropy import units as u

from ._core import CoordinateFrame

__all__ = ["SpectralFrame"]


class SpectralFrame(CoordinateFrame):
    """
    Represents Spectral Frame

    Parameters
    ----------
    axes_order : tuple or int
        A dimension in the input data that corresponds to this axis.
    reference_frame : astropy.coordinates.builtin_frames
        Reference frame (usually used with output_frame to convert to world
        coordinate objects).
    unit : str or units.Unit instance
        Spectral unit.
    axes_names : str
        Spectral axis name.
    name : str
        Name for this frame.

    """

    def __init__(
        self,
        axes_order=(0,),
        reference_frame=None,
        unit=None,
        axes_names=None,
        name=None,
        axis_physical_types=None,
    ):
        if not np.iterable(unit):
            unit = (unit,)
        unit = [u.Unit(un) for un in unit]
        pht = axis_physical_types or self._default_axis_physical_types(unit)

        super().__init__(
            naxes=1,
            axes_type="SPECTRAL",
            axes_order=axes_order,
            axes_names=axes_names,
            reference_frame=reference_frame,
            unit=unit,
            name=name,
            axis_physical_types=pht,
        )

    def _default_axis_physical_types(self, unit):
        if unit[0].physical_type == "frequency":
            return ("em.freq",)
        if unit[0].physical_type == "length":
            return ("em.wl",)
        if unit[0].physical_type == "energy":
            return ("em.energy",)
        if unit[0].physical_type == "speed":
            warnings.warn(
                "Physical type may be ambiguous. Consider "
                "setting the physical type explicitly as "
                "either 'spect.dopplerVeloc.optical' or "
                "'spect.dopplerVeloc.radio'."
            )
            return ("spect.dopplerVeloc",)
        return (f"custom:{unit[0].physical_type}",)

    @property
    def world_axis_object_classes(self):
        return {"spectral": (coord.SpectralCoord, (), {"unit": self.unit[0]})}

    @property
    def _native_world_axis_object_components(self):
        return [("spectral", 0, lambda sc: sc.to_value(self.unit[0]))]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections import OrderedDict

import numpy as np
from asdf.tags.core.ndarray import NDArrayType
from asdf_astropy.converters.transform.core import TransformConverterBase
from astropy.modeling import models
from astropy.modeling.core import Model

__all__ = ["LabelMapperConverter", "RegionsSelectorConverter"]


class LabelMapperConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/label_mapper-*",)
    types = (
        "gwcs.selector.LabelMapperArray",
        "gwcs.selector.LabelMapperDict",
        "gwcs.selector.LabelMapperRange",
        "gwcs.selector.LabelMapper",
    )

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.selector import (
            LabelMapper,
            LabelMapperArray,
            LabelMapperDict,
            LabelMapperRange,
        )

        inputs_mapping = node.get("inputs_mapping", None)
        if inputs_mapping is not None and not isinstance(
            inputs_mapping, models.Mapping
        ):
            msg = (
                "inputs_mapping must be an instanceof astropy.modeling.models.Mapping."
            )
            raise TypeError(msg)
        mapper = node["mapper"]
        atol = node.get("atol", 1e-8)
        no_label = node.get("no_label", np.nan)

        if isinstance(mapper, NDArrayType):
            if mapper.ndim != 2:
                msg = "GWCS currently only supports 2D masks."
                raise NotImplementedError(msg)
            return LabelMapperArray(mapper, inputs_mapping)
        if isinstance(mapper, Model):
            inputs = node.get("inputs")
            return LabelMapper(
                inputs, mapper, inputs_mapping=inputs_mapping, no_label=no_label
            )
        inputs = node.get("inputs", None)
        if inputs is not None:
            inputs = tuple(inputs)
        labels = mapper.get("labels")
        transforms = mapper.get("models")
        if np.iterable(labels[0]):
            labels = [tuple(label) for label in labels]
            dict_mapper = dict(zip(labels, transforms, strict=False))
            return LabelMapperRange(inputs, dict_mapper, inputs_mapping)
        dict_mapper = dict(zip(labels, transforms, strict=False))
        return LabelMapperDict(inputs, dict_mapper, inputs_mapping, atol=atol)

    def to_yaml_tree_transform(self, model, tag, ctx):
        from gwcs.selector import (
            LabelMapper,
            LabelMapperArray,
            LabelMapperDict,
            LabelMapperRange,
        )

        node = OrderedDict()
        node["no_label"] = model.no_label
        if model.inputs_mapping is not None:
            node["inputs_mapping"] = model.inputs_mapping

        if isinstance(model, LabelMapperArray):
            node["mapper"] = model.mapper
        elif isinstance(model, LabelMapper):
            node["mapper"] = model.mapper
            node["inputs"] = list(model.inputs)
        elif isinstance(model, (LabelMapperDict, LabelMapperRange)):
            if hasattr(model, "atol"):
                node["atol"] = model.atol
            mapper = OrderedDict()
            labels = list(model.mapper)

            transforms = [model.mapper[k] for k in labels]
            if np.iterable(labels[0]):
                labels = [list(label) for label in labels]
            mapper["labels"] = labels
            mapper["models"] = transforms
            node["mapper"] = mapper
            node["inputs"] = list(model.inputs)
        else:
            msg = f"Unrecognized type of LabelMapper - {model}"
            raise TypeError(msg)

        return node


class RegionsSelectorConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/regions_selector-*",)
    types = ("gwcs.selector.RegionsSelector",)

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.selector import RegionsSelector

        inputs = node["inputs"]
        outputs = node["outputs"]
        label_mapper = node["label_mapper"]
        undefined_transform_value = node["undefined_transform_value"]
        sel = node["selector"]
        sel = dict(zip(sel["labels"], sel["transforms"], strict=False))
        return RegionsSelector(
            inputs, outputs, sel, label_mapper, undefined_transform_value
        )

    def to_yaml_tree_transform(self, model, tag, ctx):
        selector = OrderedDict()
        node = OrderedDict()
        labels = list(model.selector)
        values = [model.selector[label] for label in labels]
        selector["labels"] = labels
        selector["transforms"] = values
        node["inputs"] = list(model.inputs)
        node["outputs"] = list(model.outputs)
        node["selector"] = selector
        node["label_mapper"] = model.label_mapper
        node["undefined_transform_value"] = model.undefined_transform_value
        return node
